import pytest

from steamship import Steamship
from steamship.agents.examples.document_qa_agent import ExampleDocumentQAService


@pytest.mark.usefixtures("client")
def test_qa_agent_inits(client: Steamship):
    agent = ExampleDocumentQAService(client=client)
    assert agent is not None



import logging
import uuid
from typing import List, Optional

from steamship import Block, SteamshipError, Task
from steamship.agents.llms.openai import ChatOpenAI, OpenAI
from steamship.agents.logging import AgentLogging
from steamship.agents.schema import Action, Agent, FinishAction
from steamship.agents.schema.context import AgentContext, Metadata
from steamship.agents.utils import with_llm
from steamship.invocable import PackageService, post


class AgentService(PackageService):
    """AgentService is a Steamship Package that can use an Agent, Tools, and a provided AgentContext to
    respond to user input."""

    use_llm_cache: bool
    """Whether or not to cache LLM calls (for tool selection/direct responses) by default."""

    use_action_cache: bool
    """Whether or not to cache agent Actions (for tool runs) by default."""

    def __init__(
        self,
        use_llm_cache: Optional[bool] = False,
        use_action_cache: Optional[bool] = False,
        **kwargs,
    ):
        self.use_llm_cache = use_llm_cache
        self.use_action_cache = use_action_cache
        super().__init__(**kwargs)

    ###############################################
    # Tool selection / execution
    ###############################################

    def next_action(self, agent: Agent, input_blocks: List[Block], context: AgentContext) -> Action:
        action: Action = None
        if context.llm_cache:
            action = context.llm_cache.lookup(key=input_blocks)
        if action:
            logging.info(
                f"Using cached response: calling {action.tool}.",
                extra={
                    AgentLogging.TOOL_NAME: "LLM",
                    AgentLogging.IS_MESSAGE: True,
                    AgentLogging.MESSAGE_TYPE: AgentLogging.OBSERVATION,
                    AgentLogging.MESSAGE_AUTHOR: AgentLogging.AGENT,
                },
            )
        else:
            inputs = ",".join([f"{b.as_llm_input()}" for b in input_blocks])
            logging.info(
                f"Prompting LLM with: ({inputs})",
                extra={
                    AgentLogging.TOOL_NAME: "LLM",
                    AgentLogging.IS_MESSAGE: True,
                    AgentLogging.MESSAGE_TYPE: AgentLogging.PROMPT,
                    AgentLogging.MESSAGE_AUTHOR: AgentLogging.AGENT,
                },
            )
            action = agent.next_action(context=context)
            if context.llm_cache:
                context.llm_cache.update(key=input_blocks, value=action)
        return action

    def run_action(self, agent: Agent, action: Action, context: AgentContext):
        if isinstance(action, FinishAction):
            return

        if not agent:
            raise SteamshipError(
                "Missing agent. Not able to run action on behalf of missing agent."
            )

        if context.action_cache:
            # if cache and action is cached, use it. otherwise proceed normally.
            if output_blocks := context.action_cache.lookup(key=action):
                outputs = ",".join([f"{b.as_llm_input()}" for b in output_blocks])
                logging.info(
                    f"Tool {action.tool}: ({outputs}) [cached]",
                    extra={
                        AgentLogging.TOOL_NAME: action.tool,
                        AgentLogging.IS_MESSAGE: True,
                        AgentLogging.MESSAGE_TYPE: AgentLogging.OBSERVATION,
                        AgentLogging.MESSAGE_AUTHOR: AgentLogging.AGENT,
                    },
                )
                action.output = output_blocks
                context.completed_steps.append(action)
                return

        tool = next((tool for tool in agent.tools if tool.name == action.tool), None)
        if not tool:
            raise SteamshipError(
                f"Could not find tool '{action.tool}' for action. Not able to run."
            )

        # TODO: Arrive at a solid design for the details of this structured log object
        inputs = ",".join([f"{b.as_llm_input()}" for b in action.input])
        logging.info(
            f"Running Tool {action.tool} ({inputs})",
            extra={
                AgentLogging.TOOL_NAME: action.tool,
                AgentLogging.IS_MESSAGE: True,
                AgentLogging.MESSAGE_TYPE: AgentLogging.ACTION,
                AgentLogging.MESSAGE_AUTHOR: AgentLogging.AGENT,
            },
        )
        blocks_or_task = tool.run(tool_input=action.input, context=context)
        if isinstance(blocks_or_task, Task):
            raise SteamshipError(
                "Tools return Tasks are not yet supported (but will be soon). "
                "Please use synchronous Tasks (Tools that return List[Block] for now."
            )
        else:
            outputs = ",".join([f"{b.as_llm_input()}" for b in blocks_or_task])
            logging.info(
                f"Tool {action.tool}: ({outputs})",
                extra={
                    AgentLogging.TOOL_NAME: action.tool,
                    AgentLogging.IS_MESSAGE: True,
                    AgentLogging.MESSAGE_TYPE: AgentLogging.OBSERVATION,
                    AgentLogging.MESSAGE_AUTHOR: AgentLogging.AGENT,
                },
            )
            action.output = blocks_or_task
            context.completed_steps.append(action)
            if context.action_cache:
                context.action_cache.update(key=action, value=action.output)

    def run_agent(self, agent: Agent, context: AgentContext):
        # first, clear any prior agent steps from set of completed steps
        # this will allow the agent to select tools/dispatch actions based on a new context
        context.completed_steps = []

        action = self.next_action(
            agent=agent, input_blocks=[context.chat_history.last_user_message], context=context
        )

        while not isinstance(action, FinishAction):
            self.run_action(agent=agent, action=action, context=context)
            action = self.next_action(agent=agent, input_blocks=action.output, context=context)

            # TODO: Arrive at a solid design for the details of this structured log object
            logging.info(
                f"Next Tool: {action.tool}",
                extra={
                    AgentLogging.TOOL_NAME: action.tool,
                    AgentLogging.IS_MESSAGE: False,
                    AgentLogging.MESSAGE_TYPE: AgentLogging.ACTION,
                    AgentLogging.MESSAGE_AUTHOR: AgentLogging.AGENT,
                },
            )

        context.completed_steps.append(action)
        output_text_length = 0
        if action.output is not None:
            output_text_length = sum([len(block.text or "") for block in action.output])
        logging.info(
            f"Completed agent run. Result: {len(action.output or [])} blocks. {output_text_length} total text length. Emitting on {len(context.emit_funcs)} functions."
        )
        for func in context.emit_funcs:
            logging.info(f"Emitting via function: {func.__name__}")
            func(action.output, context.metadata)

    def get_default_agent(self, throw_if_missing: bool = True) -> Optional[Agent]:
        """Return the default agent of this agent service.

        This is a helper wrapper to safeguard naming conventions that have formed.
        """
        if hasattr(self, "agent"):
            return self.agent
        elif hasattr(self, "_agent"):
            return self._agent
        else:
            if throw_if_missing:
                raise SteamshipError(
                    message="No Agent object found in the Agent Service. "
                    "Please name it either self.agent or self._agent."
                )
            else:
                return None

    def build_default_context(self, context_id: Optional[str] = None, **kwargs) -> AgentContext:
        """Build's the agent's default context.

        The provides a single place to implement (or override) the default context that will be used by endpoints
        that transports define. This allows an Agent developer to use, eg, the TelegramTransport but with a custom
        type of memory or caching.

        The returned context does not have any emit functions yet registered to it.
        """

        # AgentContexts serve to allow the AgentService to run agents
        # with appropriate information about the desired tasking.
        if context_id is None:
            context_id = uuid.uuid4()
            logging.warning(
                f"No context_id was provided; generated {context_id}. This likely means no conversational history will be present."
            )

        use_llm_cache = self.use_llm_cache
        if runtime_use_llm_cache := kwargs.get("use_llm_cache"):
            use_llm_cache = runtime_use_llm_cache

        use_action_cache = self.use_action_cache
        if runtime_use_action_cache := kwargs.get("use_action_cache"):
            use_action_cache = runtime_use_action_cache

        context = AgentContext.get_or_create(
            client=self.client,
            context_keys={"id": f"{context_id}"},
            use_llm_cache=use_llm_cache,
            use_action_cache=use_action_cache,
        )

        # Add a default LLM to the context, using the Agent's if it exists.
        llm = None
        if agent := self.get_default_agent():
            if hasattr(agent, "llm"):
                llm = agent.llm
        if llm is None:
            llm = OpenAI(client=self.client)

        context = with_llm(context=context, llm=llm)

        return context

    @post("prompt")
    def prompt(
        self, prompt: Optional[str] = None, context_id: Optional[str] = None, **kwargs
    ) -> List[Block]:
        """Run an agent with the provided text as the input."""
        prompt = prompt or kwargs.get("question")

        context = self.build_default_context(context_id, **kwargs)

        context.chat_history.append_user_message(prompt)

        # Add a default LLM
        context = with_llm(context=context, llm=ChatOpenAI(client=self.client, model_name="gpt-4"))

        # AgentServices provide an emit function hook to access the output of running
        # agents and tools. The emit functions fire at after the supplied agent emits
        # a "FinishAction".
        #
        # Here, we show one way of accessing the output in a synchronous fashion. An
        # alternative way would be to access the final Action in the `context.completed_steps`
        # after the call to `run_agent()`.
        output_blocks = []

        def sync_emit(blocks: List[Block], meta: Metadata):
            nonlocal output_blocks
            output_blocks.extend(blocks)

        context.emit_funcs.append(sync_emit)

        # Get the agent
        agent: Optional[Agent] = self.get_default_agent()
        self.run_agent(agent, context)

        # Now append the output blocks to the chat history
        # TODO: It seems like we've been going from block -> not block -> block here. Opportunity to optimize.
        for output_block in output_blocks:
            # Need to make the output blocks public here so that they can be copied to the chat history.
            # They generally need to be public anyway for the REPL to be able to show a clickable link.
            output_block.set_public_data(True)
            context.chat_history.append_assistant_message(
                text=output_block.text,
                tags=output_block.tags,
                url=output_block.raw_data_url or output_block.url or output_block.content_url,
                mime_type=output_block.mime_type,
            )

        # Return the response as a set of multi-modal blocks.
        return output_blocks



import uuid
from typing import List, Optional

from steamship import Block, Steamship, SteamshipError
from steamship.agents.mixins.transports.transport import Transport
from steamship.agents.schema import Agent, Metadata
from steamship.agents.service.agent_service import AgentService
from steamship.invocable import Config, InvocationContext, post

API_BASE = "https://api.telegram.org/bot"


class SteamshipWidgetTransport(Transport):
    """Experimental base class to encapsulate a Steamship web widget communication channel."""

    message_output: List[Block]

    def __init__(self, client: Steamship, agent_service: AgentService, agent: Agent):
        super().__init__(client=client)
        self.agent = agent
        self.agent_service = agent_service

    def instance_init(self, config: Config, context: InvocationContext):
        pass

    def _instance_deinit(self, *args, **kwargs):
        """Unsubscribe from updates."""
        pass

    def _send(self, blocks: [Block], metadata: Metadata):
        """Send a response to the client.

        TODO: Since this isn't a push, but rather an API return, we need to figure out how to model this.
        """
        pass

    @post("info")
    def info(self) -> dict:
        return {}

    def _parse_inbound(self, payload: dict, context: Optional[dict] = None) -> Optional[Block]:
        """Parses an inbound Steamship widget message."""

        message_text = payload.get("question")
        if message_text is None:
            raise SteamshipError(f"No 'question' found in Steamship widget message: {payload}")

        chat_id = payload.get("chat_session_id", "default")

        message_id = str(uuid.uuid4())

        result = Block(text=message_text)
        result.set_chat_id(str(chat_id))
        result.set_message_id(str(message_id))
        return result

    @post("answer", public=True)
    def answer(self, **payload) -> List[Block]:
        """Endpoint that implements the contract for Steamship embeddable chat widgets. This is a PUBLIC endpoint since these webhooks do not pass a token."""
        incoming_message = self.parse_inbound(payload)

        context = self.agent_service.build_default_context(context_id=incoming_message.chat_id)

        context.chat_history.append_user_message(
            text=incoming_message.text, tags=incoming_message.tags
        )
        context.emit_funcs = [self.save_for_emit]

        try:
            self.agent_service.run_agent(self.agent, context)
        except Exception as e:
            self.message_output = [self.response_for_exception(e, chat_id=incoming_message.chat_id)]

        # We don't call self.steamship_widget_transport.send because the result is the return value
        return self.message_output

    def save_for_emit(self, blocks: List[Block], metadata: Metadata):
        self.message_output = blocks



import pytest
from steamship_tests import PACKAGES_PATH
from steamship_tests.utils.deployables import deploy_package

from steamship import Steamship
from steamship.utils.url import Verb


@pytest.mark.usefixtures("client")
def test_get_param_decoding(client: Steamship):
    demo_package_path = PACKAGES_PATH / "demo_package.py"

    with deploy_package(client, demo_package_path) as (package, version, instance):
        result = instance.invoke("greet", verb=Verb.GET, name="A test with spaces")
        assert result == "Hello, A test with spaces!"



import pytest
from steamship_tests import SRC_PATH
from steamship_tests.utils.deployables import deploy_package

from steamship import Block, Steamship


@pytest.mark.usefixtures("client")
def test_fact_learner_agent_service(client: Steamship):

    agent_path = SRC_PATH / "steamship/agents/examples/fact_learner.py"
    with deploy_package(client, agent_path) as (_, _, agent):

        agent.invoke("prompt", prompt="please remember that my name is Inigo Montoya")
        agent.invoke("prompt", prompt="please remember that I am skilled swordsman")

        answer_blocks = agent.invoke("prompt", prompt="Is my name Inigo?")
        assert len(answer_blocks) == 1
        assert "yes" in Block(**answer_blocks[0]).text.lower()

        answer_blocks = agent.invoke("prompt", prompt="what do I know how to do well?")
        assert len(answer_blocks) == 1
        assert "sword" in Block(**answer_blocks[0]).text.lower()



from __future__ import annotations

import uuid
from typing import Dict, List, Optional, Union, cast

from steamship import File, MimeTypes, Steamship, SteamshipError, Task
from steamship.agents.schema.message_selectors import MessageSelector
from steamship.agents.schema.text_splitters import FixedSizeTextSplitter, TextSplitter
from steamship.base.client import Client
from steamship.data import TagKind
from steamship.data.block import Block
from steamship.data.plugin.index_plugin_instance import EmbeddingIndexPluginInstance, SearchResults
from steamship.data.tags import Tag
from steamship.data.tags.tag_constants import ChatTag, DocTag, RoleTag, TagValueKey


class ChatHistory:
    """A ChatHistory is a wrapper of a File ideal for ongoing interactions between a user and a virtual assistant.
    It also includes vector-backed storage for similarity-based retrieval."""

    file: File
    embedding_index: EmbeddingIndexPluginInstance
    text_splitter: TextSplitter

    def __init__(
        self,
        file: File,
        embedding_index: Optional[EmbeddingIndexPluginInstance],
        text_splitter: TextSplitter = None,
    ):
        """This init method is intended only for private use within the class. See `Chat.create()`"""
        self.file = file
        self.embedding_index = embedding_index
        if text_splitter is not None:
            self.text_splitter = text_splitter
        else:
            self.text_splitter = FixedSizeTextSplitter(chunk_size=300)

    @staticmethod
    def _get_existing_file(client: Client, context_keys: Dict[str, str]) -> Optional[File]:
        """Find an existing File object whose memory Tag matches the passed memory keys"""
        file_query = (
            f'kind "{TagKind.CHAT}" and name "{ChatTag.CONTEXT_KEYS}"'
            + (" and " if len(context_keys) > 0 else "")
            + " and ".join([f'value("{key}") = "{value}"' for key, value in context_keys.items()])
        )
        file = File.query(client, file_query)
        if len(file.files) == 1:
            return file.files[0]
        elif len(file.files) == 0:
            return None
        else:
            raise SteamshipError(
                "Multiple ChatHistory objects have been created in this workspace with these memory keys."
            )

    @staticmethod
    def _get_index_handle_from_file(file: File) -> str:
        for tag in file.tags:
            if tag.kind == TagKind.CHAT and tag.name == ChatTag.INDEX_HANDLE:
                return tag.value[TagValueKey.STRING_VALUE]
        raise SteamshipError(f"Could not find index handle on file with id {file.id}")

    @staticmethod
    def _get_embedding_index(client: Steamship, index_handle: str) -> EmbeddingIndexPluginInstance:
        return cast(
            EmbeddingIndexPluginInstance,
            client.use_plugin(
                plugin_handle="embedding-index",
                instance_handle=index_handle,
                config={
                    "embedder": {
                        "plugin_handle": "openai-embedder",
                        "plugin_instance-handle": "text-embedding-ada-002",
                        "fetch_if_exists": True,
                        "config": {"model": "text-embedding-ada-002", "dimensionality": 1536},
                    }
                },
                fetch_if_exists=True,
            ),
        )

    @staticmethod
    def get_or_create(
        client: Steamship,
        context_keys: Dict[str, str],
        tags: List[Tag] = None,
        searchable: bool = True,
    ) -> ChatHistory:

        file = ChatHistory._get_existing_file(client, context_keys)

        if file is None:
            tags = tags or []
            index_handle = str(uuid.uuid4())
            tags.append(Tag(kind=TagKind.DOCUMENT, name=DocTag.CHAT))
            tags.append(Tag(kind=TagKind.CHAT, name=ChatTag.CONTEXT_KEYS, value=context_keys))
            tags.append(
                Tag(
                    kind=TagKind.CHAT,
                    name=ChatTag.INDEX_HANDLE,
                    value={TagValueKey.STRING_VALUE: index_handle},
                )
            )

            blocks = []
            file = File.create(
                client=client,
                blocks=blocks,
                tags=tags,
            )
        else:
            index_handle = ChatHistory._get_index_handle_from_file(file)

        if searchable:
            embedding_index = ChatHistory._get_embedding_index(client, index_handle)
        else:
            embedding_index = None

        return ChatHistory(file, embedding_index)

    def append_message_with_role(
        self,
        text: str = None,
        role: RoleTag = RoleTag.USER,
        tags: List[Tag] = None,
        content: Union[str, bytes] = None,
        url: Optional[str] = None,
        mime_type: Optional[MimeTypes] = None,
    ) -> Block:
        """Append a new block to this with content provided by the end-user."""
        tags = tags or []
        tags.append(
            Tag(kind=TagKind.CHAT, name=ChatTag.ROLE, value={TagValueKey.STRING_VALUE: role})
        )
        block = self.file.append_block(
            text=text, tags=tags, content=content, url=url, mime_type=mime_type
        )
        if self.embedding_index is not None:
            chunk_tags = self.text_splitter.chunk_text_to_tags(
                block, kind=TagKind.CHAT, name=ChatTag.CHUNK
            )
            block.tags.extend(chunk_tags)
            self.embedding_index.insert(chunk_tags)
        return block

    def append_user_message(
        self,
        text: str = None,
        tags: List[Tag] = None,
        content: Union[str, bytes] = None,
        url: Optional[str] = None,
        mime_type: Optional[MimeTypes] = None,
    ) -> Block:
        """Append a new block to this with content provided by the end-user."""
        return self.append_message_with_role(text, RoleTag.USER, tags, content, url, mime_type)

    def append_system_message(
        self,
        text: str = None,
        tags: List[Tag] = None,
        content: Union[str, bytes] = None,
        url: Optional[str] = None,
        mime_type: Optional[MimeTypes] = None,
    ) -> Block:
        """Append a new block to this with content provided by the system, i.e., instructions to the assistant."""
        return self.append_message_with_role(text, RoleTag.SYSTEM, tags, content, url, mime_type)

    def append_assistant_message(
        self,
        text: str = None,
        tags: List[Tag] = None,
        content: Union[str, bytes] = None,
        url: Optional[str] = None,
        mime_type: Optional[MimeTypes] = None,
    ) -> Block:
        """Append a new block to this with content provided by the agent, i.e., results from the assistant."""
        return self.append_message_with_role(text, RoleTag.ASSISTANT, tags, content, url, mime_type)

    @property
    def last_user_message(self) -> Optional[Block]:
        for block in self.file.blocks[::-1]:
            if block.chat_role == RoleTag.USER:
                return block
        return None

    @property
    def last_system_message(self) -> Optional[Block]:
        for block in self.file.blocks[::-1]:
            if block.chat_role == RoleTag.SYSTEM:
                return block
        return None

    @property
    def last_agent_message(self) -> Optional[Block]:
        for block in self.file.blocks[::-1]:
            if block.chat_role == RoleTag.ASSISTANT:
                return block
        return None

    @property
    def initial_system_prompt(self) -> Optional[Block]:
        if len(self.file.blocks) > 0 and self.file.blocks[0].chat_role == RoleTag.SYSTEM:
            return self.file.blocks[0]
        else:
            return None

    def refresh(self):
        self.file.refresh()

    @property
    def tags(self) -> List[Tag]:
        return self.file.tags

    @property
    def messages(self) -> List[Block]:
        return self.file.blocks

    @property
    def client(self) -> Client:
        return self.file.client

    def select_messages(self, selector: MessageSelector) -> List[Block]:
        return selector.get_messages(self.messages)

    def search(self, text: str, k=None) -> Task[SearchResults]:
        if self.embedding_index is None:
            raise SteamshipError("This ChatHistory has no embedding index and is not searchable.")
        return self.embedding_index.search(text, k)

    def is_searchable(self) -> bool:
        return self.embedding_index is not None

    def delete_messages(self, selector: MessageSelector):
        """Delete a set of selected messages from the ChatHistory.

        If `selector == None`, no messages will be deleted.

        NOTES:
        - upon deletion, refresh() is called to ensure up-to-date history refs.
        - causes a full re-index of chat history if the history is searchable.
        """
        if selector:
            selected_messages = selector.get_messages(self.messages)
            for msg in selected_messages:
                msg.delete()

            self.refresh()
            if self.is_searchable():
                self.embedding_index.reset()
                for msg in self.messages:
                    for tag in msg.tags:
                        if tag.kind == TagKind.CHAT and tag.name == ChatTag.CHUNK:
                            # TODO(dougreid): figure out why tag.text gets lost.
                            if not tag.text:
                                tag.text = msg.text[tag.start_idx : tag.end_idx]
                            self.embedding_index.insert(tag)

        self.refresh()

    def clear(self):
        """Deletes ALL messages from the ChatHistory (including system).

        NOTE: upon deletion, refresh() is called to ensure up-to-date history refs.
        """
        for block in self.file.blocks:
            block.delete()

        if self.is_searchable():
            self.embedding_index.reset()

        self.refresh()



from abc import ABC, abstractmethod
from typing import List

from pydantic import BaseModel

from steamship import Block
from steamship.agents.schema.action import Action
from steamship.agents.schema.context import AgentContext
from steamship.agents.schema.llm import LLM, ChatLLM
from steamship.agents.schema.message_selectors import MessageSelector, NoMessages
from steamship.agents.schema.output_parser import OutputParser
from steamship.agents.schema.tool import Tool
from steamship.data.tags.tag_constants import RoleTag


class Agent(BaseModel, ABC):
    """Agent is responsible for choosing the next action to take for an AgentService.

    It uses the provided context, and a set of Tools, to decide on an action that will
    be executed by the AgentService.
    """

    tools: List[Tool]
    """Tools that can be used by the Agent in selecting the next Action."""

    message_selector: MessageSelector = NoMessages()
    """Selector of messages from ChatHistory. Used for conversation memory retrieval."""

    @abstractmethod
    def next_action(self, context: AgentContext) -> Action:
        pass


class LLMAgent(Agent):
    """LLMAgents choose next actions for an AgentService based on interactions with an LLM."""

    llm: LLM
    """The LLM to use for action selection."""

    output_parser: OutputParser
    """Utility responsible for converting LLM output into Actions"""

    @abstractmethod
    def next_action(self, context: AgentContext) -> Action:
        pass

    @staticmethod
    def messages_to_prompt_history(messages: List[Block]) -> str:
        as_strings = []
        for block in messages:
            role = block.chat_role
            if role == RoleTag.USER:
                as_strings.append(f"User: {block.text}")
            elif role == RoleTag.ASSISTANT:
                as_strings.append(f"Assistant: {block.text}")
            elif role == RoleTag.SYSTEM:
                as_strings.append(f"System: {block.text}")
            elif role == RoleTag.AGENT:
                as_strings.append(f"Agent: {block.text}")
            elif role == RoleTag.FUNCTION:
                as_strings.append(f"Function: {block.text}")
        return "\n".join(as_strings)


class ChatAgent(Agent, ABC):
    """ChatAgents choose next actions for an AgentService based on chat-based interactions with an LLM."""

    llm: ChatLLM

    output_parser: OutputParser
    """Utility responsible for converting LLM output into Actions"""



"""Answers questions with the assistance of a VectorSearch plugin."""
import logging
from typing import Any, List, Optional, Union

from steamship import Block, DocTag, Tag, Task
from steamship.agents.llms import OpenAI
from steamship.agents.logging import AgentLogging
from steamship.agents.schema import AgentContext
from steamship.agents.tools.question_answering.vector_search_tool import VectorSearchTool
from steamship.agents.utils import get_llm, with_llm
from steamship.data import TagKind
from steamship.utils.repl import ToolREPL

DEFAULT_QUESTION_ANSWERING_PROMPT = (
    "Use the following pieces of memory to answer the question at the end. "
    """If you don't know the answer, just say that you don't know, don't try to make up an answer.

{source_text}

Question: {question}

Helpful Answer:"""
)


DEFAULT_SOURCE_DOCUMENT_PROMPT = "Source Document: {text}"


class VectorSearchQATool(VectorSearchTool):
    """Tool to answer questions with the assistance of a vector search plugin."""

    name: str = "VectorSearchQATool"
    human_description: str = "Answers questions about a user. This can include personal information (names, preferences, etc.)."
    agent_description: str = (
        "Used to answer questions. ",
        "The input should be a plain text question. ",
        "The output is a plain text answer",
    )
    question_answering_prompt: Optional[str] = DEFAULT_QUESTION_ANSWERING_PROMPT
    source_document_prompt: Optional[str] = DEFAULT_SOURCE_DOCUMENT_PROMPT
    load_docs_count: int = 2

    def answer_question(self, question: str, context: AgentContext) -> List[Block]:
        embed_index = self.get_embedding_index(context.client)
        task = embed_index.search(question, k=self.load_docs_count)
        task.wait()

        source_texts = []
        source_metadata = []

        for item in task.output.items:
            if item.tag and item.tag.text:
                item_data = {"text": item.tag.text}
                source_texts.append(self.source_document_prompt.format(**item_data))
                _metadata = {}
                if item.tag.value:
                    _metadata.update(item.tag.value)
                source_metadata.append(_metadata)

        final_prompt = self.question_answering_prompt.format(
            **{"source_text": "\n".join(source_texts), "question": question}
        )

        logging.info(
            f"Tool {self.name}: sending prompt to LLM",
            extra={
                AgentLogging.TOOL_NAME: self.name,
                AgentLogging.IS_MESSAGE: True,
                AgentLogging.MESSAGE_TYPE: AgentLogging.OBSERVATION,
                AgentLogging.MESSAGE_AUTHOR: AgentLogging.TOOL,
                "prompt": final_prompt,
            },
        )
        output_blocks = get_llm(context, default=OpenAI(client=context.client)).complete(
            prompt=final_prompt
        )
        for output_block in output_blocks:
            if output_block.tags is None:
                output_block.tags = []
            output_block.tags.append(
                Tag(kind=TagKind.DOCUMENT, name=DocTag.SOURCE, value={"sources": source_metadata})
            )
        return output_blocks

    def run(self, tool_input: List[Block], context: AgentContext) -> Union[List[Block], Task[Any]]:
        """Answers questions with the assistance of an Embedding Index plugin.

        Inputs
        ------
        tool_input: List[Block]
            A list of blocks to be rewritten if text-containing.
        context: AgentContext
            The active AgentContext.

        Output
        ------
        output: List[Blocks]
            A lit of blocks containing the answers.
        """

        output = []
        for input_block in tool_input:
            if not input_block.is_text():
                continue
            for output_block in self.answer_question(input_block.text, context):
                output.append(output_block)
        return output


if __name__ == "__main__":
    tool = VectorSearchQATool()
    repl = ToolREPL(tool)

    with repl.temporary_workspace() as client:
        index = tool.get_embedding_index(client)
        index.insert([Tag(text="Ted loves apple pie."), Tag(text="The secret passcode is 1234.")])
        repl.run_with_client(
            client, context=with_llm(context=AgentContext(), llm=OpenAI(client=client))
        )



"""
THIS SHOULD BE KEPT IN SYNC WITH THE DOCS: packages/article-tagging.rst.
"""
from typing import Any, Dict, Type

from steamship import Block, File, Steamship, Tag
from steamship.invocable import Config, InvocationContext, PackageService, post


class ArticleTaggerPackage(PackageService):
    class ArticleTaggerConfig(Config):
        """Configuration required to instantiate this package."""

        labels: str  # A comma-separated list of tags to apply to articles

    def __init__(
        self, client: Steamship, config: Dict[str, Any] = None, context: InvocationContext = None
    ):
        super().__init__(client, config)

        # Instantiate a zero-shot classifier plugin
        self.classifier_instance = client.use_plugin(
            plugin_handle="zero-shot-tagger-default",
            instance_handle="my-classifier",
            config={
                "tag_kind": "tags",  # The tag.kind we want in the output
                "labels": self.config.labels,  # The labels we want to apply
                "multi_label": True,  # multi-class classification
            },
        )

    # The config_cls method allows your package to return a class
    # that defines its required configuration.
    # See Developer Reference -> Accepting Configuration
    # for more details. This package doesn't have any specific
    # required configuration, so we return the default Config object.
    @classmethod
    def config_cls(cls) -> Type[Config]:
        """Return our specific config type."""
        return cls.ArticleTaggerConfig

    # This method defines the package user's endpoint for adding content
    # The @post annotation automatically makes the method available as
    # an HTTP Post request. The name in the annotation defines the HTTP
    # route suffix, see Packages -> Package Project Structure.
    @post("add_document")
    def add_document(self, content: str, url: str) -> str:
        """Accept a new document in plaintext and start sentiment analysis"""

        # Upload the content of the file into Steamship.
        # Put the content directly into a Block, since we assume it is plaintext.
        # Create a tag with the URL so we can get it back later.
        file = File.create(
            self.client, blocks=[Block(text=content)], tags=[Tag(kind="url", name=url)]
        )

        # Tag the file with the sentiment analysis plugin
        # Using a plugin is an asynchronous call within Steamship. The
        # operation may not be complete when this method completes,
        # but that's ok. The other methods will query over whatever is
        # currently available.
        file.tag(self.classifier_instance.handle)

        return file.handle

    @staticmethod
    def _find_url(file: File):
        for tag in file.tags:
            if tag.kind == "url":
                return tag.name

    @post("documents_by_tag")
    def documents_by_tag(self, tag: str, threshold: float = 0.7) -> [str]:
        """Query the stored documents for tagged articles"""

        # Query our documents for Positive sentiment tags
        matching_files = File.query(
            self.client, f'kind "tags" and name "{tag}" and value("score") > {threshold}'
        ).files

        return [self._find_url(file) for file in matching_files]



import pytest
import requests
from assets.packages.transports.mock_slack_api import MockSlackApi
from steamship_tests import PACKAGES_PATH
from steamship_tests.utils.deployables import deploy_package

from steamship import File, Steamship

config_template = {
    "bot_token": {"type": "string"},
    "api_base": {"type": "string"},
    "slack_api_base": {"type": "string"},
}


@pytest.mark.usefixtures("client")
def test_slack(client: Steamship):
    mock_slack_api_path = PACKAGES_PATH / "transports" / "mock_slack_api.py"
    transport_agent_path = PACKAGES_PATH / "transports" / "test_transports_agent.py"

    with deploy_package(client, mock_slack_api_path) as (_, _, mock_chat_api):

        # Set the slack_api_base to the mock Slack API we just deployed.
        # This will allow us to test for the proper receipt of messages.
        instance_config = {
            "slack_api_base": mock_chat_api.invocation_url,
            "bot_token": "",
            "api_base": "",
        }

        with deploy_package(
            client,
            transport_agent_path,
            version_config_template=config_template,
            instance_config=instance_config,
        ) as (_, _, agent_instance):
            # Set the bot token
            is_token_set_no = agent_instance.invoke("is_slack_token_set")
            assert is_token_set_no is False

            agent_instance.invoke("set_slack_access_token", token="")  # noqa: S106
            is_token_set_true = agent_instance.invoke("is_slack_token_set")
            assert is_token_set_true is True

            # test sending another message; this has been a problem before
            respond_method = "slack_event"

            # Set the response URL
            response_url = agent_instance.invocation_url
            if not response_url.endswith("/"):
                response_url = response_url + "/"
            response_url = response_url + respond_method

            # Test sending messages (without auth)
            response = requests.post(
                url=response_url,
                json=MockSlackApi.generate_inbound_webhook_body("a test"),
            )
            assert response.ok
            files = File.query(client, f'kind "{MockSlackApi.TEXT_MESSAGE_TAG}"').files
            assert len(files) == 1
            allowed_responses = ["Response to: a test", "Response to: a test".replace(" ", "+")]
            assert (
                files[0].tags[0].name in allowed_responses
            )  # bug somewhere - results being url encoded in some envs
            assert files[0].tags[0].value == {MockSlackApi.CHAT_ID_KEY: "1"}

            # test sending another message; this has been a problem before
            agent_instance.invoke(
                respond_method, **MockSlackApi.generate_inbound_webhook_body("another test")
            )
            files = File.query(client, f'kind "{MockSlackApi.TEXT_MESSAGE_TAG}"').files
            assert len(files) == 2
            allowed_responses.extend(
                ["Response to: another test", "Response to: another test".replace(" ", "+")]
            )
            for file in files:
                assert (
                    file.tags[0].name in allowed_responses
                )  # bug somewhere - results being url encoded in some envs
                assert file.tags[0].value == {MockSlackApi.CHAT_ID_KEY: "1"}

            # Test the agent sending a "photo"
            agent_instance.invoke(
                respond_method, **MockSlackApi.generate_inbound_webhook_body("image")
            )
            files = File.query(client, f'kind "{MockSlackApi.PHOTO_MESSAGE_TAG}"').files
            assert len(files) == 1
            assert files[0].tags[0].value.get(MockSlackApi.CHAT_ID_KEY) == "1"
            assert client.config.api_base in files[0].tags[0].value.get(MockSlackApi.PHOTO_KEY)

            # Test the agent sending "audio"
            agent_instance.invoke(
                respond_method, **MockSlackApi.generate_inbound_webhook_body("audio")
            )
            files = File.query(client, f'kind "{MockSlackApi.AUDIO_MESSAGE_TAG}"').files
            assert len(files) == 1
            assert files[0].tags[0].value.get(MockSlackApi.CHAT_ID_KEY) == "1"
            assert client.config.api_base in files[0].tags[0].value.get(MockSlackApi.AUDIO_KEY)

            # Test the agent sending a "video"
            agent_instance.invoke(
                respond_method, **MockSlackApi.generate_inbound_webhook_body("video")
            )
            files = File.query(client, f'kind "{MockSlackApi.VIDEO_MESSAGE_TAG}"').files
            assert len(files) == 1
            assert files[0].tags[0].value.get(MockSlackApi.CHAT_ID_KEY) == "1"
            assert client.config.api_base in files[0].tags[0].value.get(MockSlackApi.VIDEO_KEY)



from typing import Optional, cast

from steamship import Block, DocTag, File, Steamship, Tag
from steamship.data import TagKind, TagValueKey
from steamship.data.plugin.index_plugin_instance import EmbeddingIndexPluginInstance, SearchResults
from steamship.invocable import post
from steamship.invocable.package_mixin import PackageMixin
from steamship.utils.file_tags import update_file_status
from steamship.utils.text_chunker import chunk_text

DEFAULT_EMBEDDING_INDEX_CONFIG = {
    "embedder": {
        "plugin_handle": "openai-embedder",
        "plugin_instance_handle": "text-embedding-ada-002",
        "fetch_if_exists": True,
        "config": {"model": "text-embedding-ada-002", "dimensionality": 1536},
    }
}

DEFAULT_EMBEDDING_INDEX_HANDLE = "default-embedding-index"


class IndexerMixin(PackageMixin):
    """Provides endpoints for easy Indexing of blockified files."""

    client: Steamship
    context_window_size: int
    context_window_overlap: int
    embedding_index_config: dict

    def __init__(
        self,
        client: Steamship,
        embedder_config: dict = None,
        context_window_size: int = 200,
        context_window_overlap: int = 50,
    ):
        self.client = client
        self.context_window_size = context_window_size
        self.context_window_overlap = context_window_overlap
        self.embedding_indexes = {}
        self.embedding_index_config = embedder_config or DEFAULT_EMBEDDING_INDEX_CONFIG

    def _get_page(self, block: Block) -> Optional[str]:
        """Return the page_id from the block if it exists."""
        page_id = None
        for tag in block.tags:
            if tag.name == DocTag.PAGE:
                page_num = tag.value.get(TagValueKey.NUMBER_VALUE)
                if page_num is not None:
                    page_id = page_num
        return page_id

    def _get_index(self, index_handle: Optional[str] = None) -> EmbeddingIndexPluginInstance:
        """Return the page_id from the block if it exists."""
        handle = index_handle or DEFAULT_EMBEDDING_INDEX_HANDLE
        if handle in self.embedding_indexes:
            return self.embedding_indexes[handle]

        self.embedding_indexes[handle] = cast(
            EmbeddingIndexPluginInstance,
            self.client.use_plugin(
                plugin_handle="embedding-index",
                instance_handle=handle,
                config=self.embedding_index_config,
                fetch_if_exists=True,
            ),
        )
        return self.embedding_indexes[handle]

    @post("/index_text")
    def index_text(
        self, text: str, metadata: Optional[dict] = None, index_handle: Optional[str] = None
    ) -> bool:
        """Load text into an embedding index.

        Optional arguments:
        - index_handle (uses your default index if blank)
        - metadata (returned on embedding results for source attribution)
        """
        tags = []
        for chunk in chunk_text(
            text, chunk_size=self.context_window_size, chunk_overlap=self.context_window_overlap
        ):
            tags.append(Tag(text=chunk, value=metadata))
        self._get_index(index_handle).insert(tags)
        return True

    def _index_block(
        self, block: Block, metadata: Optional[dict] = None, index_handle: Optional[str] = None
    ):
        page_id = self._get_page(block)
        _metadata = {}
        if metadata:
            _metadata.update(metadata)

        _metadata.update(
            {
                "file_id": block.file_id,
                "block_id": block.id,
                "page": page_id,
            }
        )

        return self.index_text(block.text, metadata=_metadata, index_handle=index_handle)

    @post("/index_block")
    def index_block(
        self, block_id: str, metadata: Optional[dict] = None, index_handle: Optional[str] = None
    ):
        """Load a Steamship Block into an embedding index.

        Optional arguments:
        - index_handle (uses your default index if blank)
        - metadata (returned on embedding results for source attribution)
        """
        block = Block.get(self.client, _id=block_id)
        page_id = self._get_page(block)
        _metadata = {}
        _metadata.update(metadata)
        _metadata.update(
            {
                "file_id": block.file_id,
                "block_id": block.id,
                "page": page_id,
            }
        )

        return self.index_text(block.text, metadata=_metadata, index_handle=index_handle)

    @post("/index_file")
    def index_file(
        self, file_id: str, metadata: Optional[dict] = None, index_handle: Optional[str] = None
    ) -> bool:
        """Load a Steamship File into an embedding index.

        Optional arguments:
        - index_handle (uses your default index if blank)
        - metadata (returned on embedding results for source attribution)
        """
        file = File.get(self.client, _id=file_id)
        update_file_status(self.client, file, "Indexing")

        _metadata = {}
        if file.mime_type:
            _metadata["mime_type"] = file.mime_type

        for tag in file.tags or []:
            if tag.kind == TagKind.DOCUMENT and tag.name == DocTag.TITLE:
                if title := tag.value.get(TagValueKey.STRING_VALUE):
                    _metadata["title"] = title

        if metadata:
            _metadata.update(metadata)

        for block in file.blocks or []:
            self._index_block(block, metadata=_metadata, index_handle=index_handle)

        update_file_status(self.client, file, "Indexed")
        return True

    @post("/search_index")
    def search_index(
        self, query: str, index_handle: Optional[str] = None, k: int = 5
    ) -> SearchResults:
        """Search an embedding index.

        Optional arguments:
        - index_handle (uses your default index if blank)
        """
        index = self._get_index(index_handle)
        task = index.search(query, k)
        return task.wait()



from typing import List, Optional

from steamship import Block, File, PluginInstance, Steamship
from steamship.agents.schema import LLM, ChatLLM, Tool

PLUGIN_HANDLE = "gpt-4"
DEFAULT_MAX_TOKENS = 256


class OpenAI(LLM):
    """LLM that uses Steamship's OpenAI plugin to generate completions.

    NOTE: By default, this LLM uses the `gpt-3.5-turbo` model. Valid model
    choices are `gpt-3.5-turbo` and `gpt-4`.
    """

    generator: PluginInstance
    client: Steamship

    def __init__(
        self, client, model_name: str = "gpt-3.5-turbo", temperature: float = 0.4, *args, **kwargs
    ):
        """Create a new instance.

        Valid model names are:
         - gpt-4
         - gpt-3.5-turbo

        Supported kwargs include:
        - `max_tokens` (controls the size of LLM responses)
        """
        client = client
        max_tokens = DEFAULT_MAX_TOKENS
        if "max_tokens" in kwargs:
            max_tokens = kwargs["max_tokens"]

        generator = client.use_plugin(
            PLUGIN_HANDLE,
            config={"model": model_name, "temperature": temperature, "max_tokens": max_tokens},
        )
        super().__init__(client=client, generator=generator, *args, **kwargs)

    def complete(self, prompt: str, stop: Optional[str] = None, **kwargs) -> List[Block]:
        """Completes the prompt, respecting the supplied stop sequence.

        Supported kwargs include:
        - `max_tokens` (controls the size of LLM responses)
        """
        options = {}
        if stop:
            options["stop"] = stop

        if "max_tokens" in kwargs:
            options["max_tokens"] = kwargs["max_tokens"]

        action_task = self.generator.generate(text=prompt, options=options)
        action_task.wait()
        return action_task.output.blocks


class ChatOpenAI(ChatLLM, OpenAI):
    """ChatLLM that uses Steamship's OpenAI plugin to generate chat completions."""

    def __init__(self, client, model_name: str = "gpt-4-0613", *args, **kwargs):
        """Create a new instance.

        Valid model names are:
         - gpt-4
         - gpt-4-0613

        Supported kwargs include:
        - `max_tokens` (controls the size of LLM responses)
        """
        super().__init__(client=client, model_name=model_name, *args, **kwargs)

    def chat(self, messages: List[Block], tools: Optional[List[Tool]], **kwargs) -> List[Block]:
        """Sends chat messages to the LLM with functions from the supplied tools in a side-channel.

        Supported kwargs include:
        - `max_tokens` (controls the size of LLM responses)
        """

        temp_file = File.create(client=self.client, blocks=messages)

        options = {}
        if len(tools) > 0:
            functions = []
            for tool in tools:
                functions.append(tool.as_openai_function())
            options["functions"] = functions

        if "max_tokens" in kwargs:
            options["max_tokens"] = kwargs["max_tokens"]

        tool_selection_task = self.generator.generate(input_file_id=temp_file.id, options=options)
        tool_selection_task.wait()
        return tool_selection_task.output.blocks



import json
import logging
import signal
import sys
import time
from os import getenv, path
from typing import Optional

import click

import steamship
from steamship import PackageInstance, Steamship, SteamshipError, Workspace
from steamship.base.configuration import DEFAULT_WEB_BASE, Configuration
from steamship.cli.create_instance import (
    config_str_to_dict,
    create_instance,
    load_manifest,
    set_unset_params,
)
from steamship.cli.deploy import (
    PackageDeployer,
    PluginDeployer,
    bundle_deployable,
    update_config_template,
)
from steamship.cli.local_server.server import SteamshipHTTPServer
from steamship.cli.manifest_init_wizard import manifest_init_wizard
from steamship.cli.requirements_init_wizard import requirements_init_wizard
from steamship.cli.ship_spinner import ship_spinner
from steamship.cli.utils import find_api_py, get_api_module
from steamship.data.manifest import DeployableType, Manifest
from steamship.data.user import User
from steamship.invocable.dev_logging_handler import DevelopmentLoggingHandler
from steamship.invocable.lambda_handler import get_class_from_module
from steamship.utils.repl import HttpREPL


@click.group()
def cli():
    pass


def initialize(suppress_message: bool = False):
    logging.root.setLevel(logging.FATAL)
    if not suppress_message:
        click.echo(f"Steamship Python CLI version {steamship.__version__}")


def initialize_and_get_client_and_prep_project():
    initialize()
    client = None
    try:
        client = Steamship()
    except SteamshipError as e:
        click.secho(e.message, fg="red")
        click.get_current_context().abort()

    user = User.current(client)
    if path.exists("steamship.json"):
        manifest = Manifest.load_manifest()
    else:
        manifest = manifest_init_wizard(client)
        manifest.save()

    if not path.exists("requirements.txt"):
        requirements_init_wizard()

    update_config_template(manifest)

    return client, user, manifest


@click.command()
def login():
    """Log in to Steamship, creating ~/.steamship.json"""
    initialize()
    click.echo("Logging into Steamship.")
    if sys.argv[1] == "login":
        if Configuration.default_config_file_has_api_key():
            overwrite = click.confirm(
                text="You already have an API key in your .steamship.json file.  Do you want to remove it and login?",
                default=False,
            )
            if not overwrite:
                sys.exit(0)
            Configuration.remove_api_key_from_default_config()

        # Carry on with login
        client = Steamship()
        user = User.current(client)
        click.secho(f"🚢🚢🚢 Hooray! You're logged in with user handle: {user.handle} 🚢🚢🚢", fg="green")


@click.command()
def ships():
    """Ship some ships"""
    initialize()
    click.secho("Here are some ships:", fg="cyan")
    with ship_spinner():
        time.sleep(5)
    click.secho()


def _run_ngrok(local_port: int) -> str:
    """Create an NGROK URL directed at `local_port`."""
    try:
        from pyngrok import ngrok
    except BaseException:
        click.secho("⚠️ Public API: Unable to start ngrok. Please either:")
        click.secho("   - Install pyngrok via `pip install pyngrok`")
        click.secho("   - Use the --no_ngrok flag")
        click.secho("   NGROK is only necessary if you wish to debug Telegram or Slack locally.")
        exit(1)
    try:
        http_tunnel = ngrok.connect(local_port, bind_tls=True)
    except BaseException:
        click.secho(f"⚠️ Public API: Unable to bind ngrok to port {local_port}")
        click.secho("   - Try running with a different port via the  --port flag.")
        exit(1)

    ngrok_api_url = http_tunnel.public_url
    return ngrok_api_url


def _run_local_server(
    local_port: int,
    instance_handle: Optional[str] = None,
    config: Optional[str] = None,
    workspace: Optional[str] = None,
    base_url: Optional[
        str
    ] = None,  # If provided, will override the default calculation, eg for ngrok
) -> str:
    """Start the local API Server."""
    logging.info(
        f"Starting local server. port={local_port}, instance_handle={instance_handle}, workspace={workspace}"
    )

    path = find_api_py()
    api_module = get_api_module(path)
    invocable_class = get_class_from_module(api_module)

    # Use the provided base url (e.g. from NGROK) or default to localhost
    _base_url = base_url or "http://localhost"

    logging.info(f"Local server base URL (for PackageService configuration) is {_base_url}")

    if not invocable_class:
        logging.error("Local server startup unable to find Steamship service.")
        click.secho("⚠️ Local API: Unable to find Steamship service. Please:")
        click.secho(
            "   - Check to see that you have an api.py file containing an AgentService or PackageService "
        )
        exit(1)

    manifest = load_manifest()

    if not manifest:
        logging.error("Local server startup unable to find Steamship manifest.")
        click.secho("⚠️ Local API: Unable to find your steamship.json file")
        exit(1)

    invocable_config, is_file = config_str_to_dict(config)
    set_unset_params(config, invocable_config, is_file, manifest)

    server = SteamshipHTTPServer(
        invocable_class,
        base_url=_base_url,
        port=local_port,
        invocable_handle=manifest.handle,
        invocable_version_handle=manifest.version,
        invocable_instance_handle=instance_handle,
        config=invocable_config,
        workspace=workspace,
    )
    try:
        server.start()
    except SteamshipError as e:
        click.secho(f"⚠️ Local API: {e.message}")
        if "validation error" in e.message:
            click.secho(
                "💡 Suggestion: Create a configuration file called config.json containing the required "
            )
            click.secho(
                "               configuration fields in your agent's configuration. Then run again with"
            )
            click.secho("               ship run local -c config.json")
            exit(-1)
    except BaseException as e:
        click.secho(f"⚠️ Local API: {e}")
        exit(-1)

    logging.info("Local server running.")

    def on_exit(signum, frame):
        logging.info("Local server shutting down.")
        click.secho("Shutting down server.")
        server.stop()
        click.secho("Shut down.")
        logging.info("Local server shut down complete.")
        exit(1)

    signal.signal(signal.SIGINT, on_exit)

    local_api_url = f"http://localhost:{server.port}"

    logging.info(f"Local server address: {local_api_url}")

    return local_api_url


def _run_web_interface(base_url: str) -> str:
    web_base = DEFAULT_WEB_BASE
    logging.info(f"Starting web interface. web_base={web_base} and base_url={base_url}")

    try:
        config = Configuration()
        web_base = config.web_base
    except BaseException:
        click.secho("⚠️ Web UI:  Unable to find Steamship Configuration. Please:")
        click.secho(
            "   - Run `ship login` to make sure you have Steamship credentials in your environment."
        )

    # Guarantee one and only one trailing /
    if not web_base.endswith("/"):
        web_base = f"{web_base}/"

    web_url = f"{web_base}debug?endpoint={base_url}/prompt"
    logging.info(f"Web interface url is: {web_url}")

    return web_url


def register_locally_running_package_with_engine(
    client: Steamship,
    ngrok_api_url: str,
    package_handle: str,
    manifest: Manifest,
    config: Optional[str] = None,
) -> PackageInstance:
    """Registers the locally running package with the Steamship Engine."""

    # Register the Instance in the Engine
    invocable_config, is_file = config_str_to_dict(config)
    set_unset_params(config, invocable_config, is_file, manifest)
    package_instance = PackageInstance.create_local_development_instance(
        client,
        local_development_url=ngrok_api_url,
        package_handle=package_handle,
        config=invocable_config,
    )
    return package_instance


def serve_local(  # noqa: C901
    port: int = 8443,
    no_ngrok: Optional[bool] = False,
    no_repl: Optional[bool] = False,
    no_ui: Optional[bool] = False,
    config: Optional[str] = None,
    workspace: Optional[str] = None,
):
    """Serve the invocable on localhost. Useful for debugging locally."""
    dev_logging_handler = DevelopmentLoggingHandler.init_and_take_root()

    click.secho("Running your project...\n")

    client, user, manifest = initialize_and_get_client_and_prep_project()

    if workspace:
        workspace_obj = Workspace.get(client, handle=workspace)
    else:
        workspace_obj = Workspace.get(client)
        workspace = workspace_obj.handle

    # Make sure we're running a package.
    if manifest.type != DeployableType.PACKAGE:
        click.secho(
            f"⚠️ Must run `ship serve local` in a folder with a Steamship Package. Found: {manifest.type}"
        )
        exit(-1)

    # Make sure we have a package name -- this allows us to register the running copy with the engine.
    deployer = PackageDeployer()
    deployable = deployer.create_or_fetch_deployable(client, user, manifest)

    # Report the logs output file.
    click.secho(f"📝 Log file:   {dev_logging_handler.log_filename}")

    # Start the NGROK connection
    ngrok_api_url = None
    public_api_url = None

    if not no_ngrok:
        ngrok_api_url = _run_ngrok(port)

        # It requires a trailing slash
        if ngrok_api_url[-1] != "/":
            ngrok_api_url = ngrok_api_url + "/"

        registered_instance = register_locally_running_package_with_engine(
            client=client,
            ngrok_api_url=ngrok_api_url,
            package_handle=deployable.handle,
            manifest=manifest,
            config=config,
        )

        # Notes:
        #  1. registered_instance.invocation_url is the NGROK URL, not the Steamship Proxy URL.
        #  2. The public_api_url should still be NGROK, not the Proxy. The local server emulates the Proxy and
        #     the Proxy blocks this kind of development traffic.

        public_api_url = ngrok_api_url
        click.secho(f"🌎 Public API: {public_api_url}")

    # Start the local API Server. This has to happen after NGROK because the port & url need to be plummed.
    try:
        local_api_url = _run_local_server(
            local_port=port,
            instance_handle=registered_instance.handle,
            config=config,
            workspace=workspace,
            base_url=public_api_url,
        )
    except BaseException as e:
        click.secho("⚠️ Local API:  Unable to start local server.")
        click.secho(e)
        exit(-1)

    if local_api_url[-1] != "/":
        local_api_url = local_api_url + "/"

    if local_api_url:
        click.secho(f"🌎 Local API:  {local_api_url}")
    else:
        click.secho("⚠️ Local API:  Unable to start local server.")
        exit(-1)

    # Start the web UI
    if not no_ui:
        web_url = _run_web_interface(public_api_url or local_api_url)
        if web_url:
            click.secho(f"🌎 Web UI:     {web_url}")

    if no_repl:
        while True:
            # We need to make sure the thread doesn't exit.
            time.sleep(1)
    else:
        click.secho("\n💬 Interactive REPL below. Type to interact.\n")
        prompt_url = f"{local_api_url or public_api_url}prompt"
        repl = HttpREPL(prompt_url=prompt_url, dev_logging_handler=dev_logging_handler)
        repl.run()


@click.command()
@click.option(
    "--port",
    "-p",
    type=int,
    default=8443,
    help="Port to host the server on.",
)
@click.option(
    "--instance_handle",
    "-h",
    type=str,
    default=None,
    help="Handle of the package or plugin instance being hosted.",
)
@click.option(
    "--no-ngrok",
    is_flag=True,
    default=False,
    help="Don't attempt to attach to ngrok.",
)
@click.option(
    "--no-ui",
    is_flag=True,
    default=False,
    help="Don't attempt to attach to a web UI.",
)
@click.option(
    "--no-repl",
    is_flag=True,
    default=False,
    help="Don't start a console REPL.",
)
@click.option(
    "--config",
    "-c",
    type=str,
    required=False,
    help="Instance configuration. May be inline JSON or a path to a file. If not specified, "
    "an empty configuration dictionary will be passed to the instance.",
)
@click.option(
    "--workspace",
    "-w",
    required=False,
    type=str,
    help="Workspace handle. The new instance will be created in this workspace. If not specified, "
    "the default workspace will be used.",
)
@click.argument(
    "environment", required=True, type=click.Choice(["local", "remote"], case_sensitive=False)
)
@click.pass_context
def run(
    ctx,
    environment: str,
    port: int = 8080,
    instance_handle: Optional[str] = None,
    no_ngrok: Optional[bool] = False,
    no_ui: Optional[bool] = False,
    no_repl: Optional[bool] = False,
    config: Optional[str] = None,
    workspace: Optional[str] = None,
):
    """Serve your invocable locally or on prod"""
    if environment == "local":
        serve_local(
            port=port,
            no_ngrok=no_ngrok,
            no_repl=no_repl,
            no_ui=no_ui,
            config=config,
            workspace=workspace,
        )
    else:
        if click.confirm("Do you want to deploy a new version first?", default=False):
            ctx.invoke(deploy)
        ctx.invoke(
            create_instance, workspace=workspace, instance_handle=instance_handle, config=config
        )


@click.command()
def deploy():
    """Deploy the package or plugin in this directory"""
    client, user, manifest = initialize_and_get_client_and_prep_project()

    deployable_type = manifest.type

    deployer = None
    if deployable_type == DeployableType.PACKAGE:
        deployer = PackageDeployer()
    elif deployable_type == DeployableType.PLUGIN:
        deployer = PluginDeployer()
    else:
        click.secho("Deployable must be of type package or plugin.", fg="red")
        click.get_current_context().abort()

    deployable = deployer.create_or_fetch_deployable(client, user, manifest)

    click.echo("Bundling content... ", nl=False)
    bundle_deployable(manifest)
    click.echo("Done. 📦")

    _ = deployer.create_version(client, manifest, deployable.id)

    thing_url = f"{client.config.web_base}{deployable_type.value}s/{manifest.handle}"
    click.echo(
        f"Deployment was successful. View and share your new {deployable_type.value} here:\n\n{thing_url}\n"
    )

    # Common error conditions:
    # - Package/plugin handle already taken. [handled; asks user for new]
    # - Version handle already deployed. [handled; asks user for new]
    # - Bad parameter configuration. [mitigated by deriving template from Config object]
    # - Package content fails health check (ex. bad import) [Error caught while checking config object]


@click.command()
def info():
    """Displays information about the current Steamship user.

    This is useful to help users (in a support or hackathon context) test whether they have configured their
    Steamship environment correctly.
    """
    initialize()
    click.echo("\nSteamship Client Info\n=====================\n")

    if Configuration.default_config_file_has_api_key() or getenv("STEAMSHIP_API_KEY", None):
        # User is logged in!
        client = None
        try:
            client = Steamship()
        except BaseException:
            click.secho("Incorrect API key or network error.\n", fg="red")
            click.secho(
                "Your Steamship API Key is set, but we were unable to use it to fetch your account information.\n"
            )
            click.secho("- If you are on your own computer, run `ship login` to login.")
            click.secho(
                "- If you are in Replit, add the STEAMSHIP_API_KEY secret, then close and re-open this shell.\n"
            )
            return

        try:
            user = User.current(client)
            click.echo(f"User handle: {user.handle}")
            click.echo(f"User ID:     {user.id}")
            click.echo(f"Profile:     {client.config.profile}")
            click.echo("\nReady to ship! 🚢🚢🚢\n")

        except BaseException:
            click.secho("Incorrect API key or network error.\n", fg="red")
            click.secho(
                "Your Steamship API Key is set, but we were unable to use it to fetch your account information.\n"
            )
            click.secho("- If you are on your own computer, run `ship login` to login.")
            click.secho(
                "- If you are in Replit, add the STEAMSHIP_API_KEY secret, then close and re-open this shell.\n"
            )
    else:
        click.secho("You are not logged in.\n")
        click.secho("- If you are on your own computer, run `ship login` to login.")
        click.secho(
            "- If you are in Replit, add the STEAMSHIP_API_KEY secret, then close and re-open this shell.\n"
        )


@click.command()
@click.option(
    "--workspace",
    "-w",
    required=True,
    type=str,
    help="Workspace handle used for scoping logs request. All requests MUST be scoped by workspace.",
)
@click.option(
    "--offset",
    "-o",
    default=0,
    type=int,
    help="Starting index from sorted logs to return a chunk. Used for paging. Defaults to 0.",
)
@click.option(
    "--number",
    "-n",
    default=50,
    type=int,
    help="Number of logs to return in a single batch. Defaults to 50.",
)
@click.option(
    "--package",
    "-p",
    type=str,
    help="Package handle. Used to filter logs returend to a specific package (across all instances).",
)
@click.option(
    "--instance",
    "-i",
    type=str,
    help="Instance handle. Used to filter logs returned to a specific instance of a package.",
)
@click.option(
    "--version",
    "-v",
    type=str,
    help="Version handle. Used to filter logs returned to a specific version of a package.",
)
@click.option(
    "--path",
    "request_path",
    type=str,
    help="Path invoked by a client operation. Used to filter logs returned to a specific invocation path.",
)
@click.option(
    "--with-fields",
    "-f",
    "field_values",
    type=str,
    help="Dictionary of log field values (format: key1=value1,key2=value2,...). Used to filter logs returned.",
)
def logs(
    workspace: str,
    offset: int,
    number: int,
    package: Optional[str] = None,
    instance: Optional[str] = None,
    version: Optional[str] = None,
    request_path: Optional[str] = None,
    field_values: Optional[str] = None,
):
    """Retrieve logs within a workspace."""

    initialize(suppress_message=True)
    client = None
    try:
        client = Steamship(workspace=workspace)
    except SteamshipError as e:
        raise click.UsageError(message=e.message)

    value_dict = {}
    if field_values:
        try:
            for item in field_values.split(","):
                key, value = item.split("=")
                value_dict[key] = value
        except ValueError:
            raise click.UsageError(
                message="Invalid dictionary format for fields. Please provide a dictionary in the "
                "format: key1=value1,key2=value2,..."
            )

    click.echo(
        json.dumps(
            client.logs(offset, number, package, instance, version, request_path, value_dict)
        )
    )


def _exit_if_not_proceed(skip: bool, prompt: str):
    if skip:
        return

    click.secho("⚠️ Deletion is a destructive operation. It is not recoverable. ⚠️", fg="red")
    click.secho(prompt, fg="red")
    confirm = click.confirm("Proceed?", default=False)
    if not confirm:
        exit(0)


@click.command()
@click.option(
    "--workspace",
    "-w",
    required=True,
    type=str,
    help="Workspace handle.",
)
@click.option(
    "--instance",
    "-i",
    type=str,
    help="Instance handle.",
)
@click.option("--yes", is_flag=True)
def delete(
    workspace: str,
    instance: Optional[str] = None,
    yes: Optional[bool] = False,
):
    """Deletes Steamship workspaces and instances, based on provided fields.

    NOTE: If `instance` is not specified, the `workspace` will be deleted.
    """
    initialize(suppress_message=True)
    client = None
    try:
        client = Steamship(workspace=workspace)
    except SteamshipError as e:
        raise click.UsageError(message=e.message)

    if not instance:
        # delete workspace
        _exit_if_not_proceed(
            skip=yes,
            prompt=f"This will delete workspace '{workspace}' and all data contained within.",
        )
        wspace = client.get_workspace()
        click.secho(f"Deleting workspace '{workspace}'... ", nl=False, fg="green")
        wspace.delete()
        click.secho("Done.", fg="green")
        return

    if instance:
        # delete instance
        _exit_if_not_proceed(
            skip=yes, prompt=f"This will delete instance '{instance}' in workspace '{workspace}'"
        )
        try:
            pkg_inst = PackageInstance.get(client=client, handle=instance)
            click.secho(
                f"Deleting instance '{instance}' in workspace '{workspace}'... ",
                nl=False,
                fg="green",
            )
            pkg_inst.delete()
            click.secho("Done.", fg="green")
            return
        except SteamshipError as e:
            click.secho(
                f"⚠️ Failed to delete instance '{instance}' in workspace '{workspace}': {e.message}",
                fg="red",
            )
            exit(1)


cli.add_command(login)
cli.add_command(deploy)
cli.add_command(info)
cli.add_command(deploy, name="it")
cli.add_command(ships)
cli.add_command(logs)
cli.add_command(run)
cli.add_command(create_instance, name="use")
cli.add_command(delete)



from typing import Any, Dict, Optional, Type

from steamship import Block, File, MimeTypes, Steamship
from steamship.agents.mixins.transports.slack import SlackTransport, SlackTransportConfig
from steamship.agents.mixins.transports.steamship_widget import SteamshipWidgetTransport
from steamship.agents.mixins.transports.telegram import TelegramTransport, TelegramTransportConfig
from steamship.agents.schema import Action, Agent, AgentContext, FinishAction
from steamship.agents.service.agent_service import AgentService
from steamship.invocable import Config, InvocationContext


class TestTransportsAgentConfig(Config):
    bot_token: Optional[str] = ""
    api_base: Optional[str] = ""
    slack_api_base: Optional[str] = ""


class TestAgent(Agent):
    client: Steamship

    def __init__(self, client: Steamship):
        super(TestAgent, self).__init__(client=client, tools=[])
        self.client = client

    def next_action(self, context: AgentContext) -> Action:
        """Helps us test the transport by controlling what it will return"""
        input_text = context.chat_history.last_user_message.text
        if input_text == "image":
            output_file = File.create(self.client, blocks=[])
            output_block = Block.create(
                self.client,
                content="some image bytes",
                mime_type=MimeTypes.PNG,
                file_id=output_file.id,
                public_data=True,
            )
        elif input_text == "audio":
            output_file = File.create(self.client, blocks=[])
            output_block = Block.create(
                self.client,
                content="some audio bytes",
                mime_type=MimeTypes.WAV,
                file_id=output_file.id,
                public_data=True,
            )
        elif input_text == "video":
            output_file = File.create(self.client, blocks=[])
            output_block = Block.create(
                self.client,
                content="some video bytes",
                mime_type=MimeTypes.MP4_VIDEO,
                file_id=output_file.id,
                public_data=True,
            )
        else:
            output_block = Block(text=f"Response to: {input_text}")
        return FinishAction(output=[output_block], context=context)


class TestTransportsAgentService(AgentService):

    config: TestTransportsAgentConfig
    agent: Agent

    USED_MIXIN_CLASSES = [
        TelegramTransport,
        SteamshipWidgetTransport,
        SlackTransport,
        # TODO: Future Transport authors: add your transport here.
    ]

    def __init__(
        self, client: Steamship, config: Dict[str, Any] = None, context: InvocationContext = None
    ):
        super().__init__(client=client, config=config, context=context)
        self.agent = TestAgent(client)

        # Including the web widget transport on the telegram test
        # agent to make sure it doesn't interfere
        self.add_mixin(
            SteamshipWidgetTransport(client=client, agent_service=self, agent=self.agent)
        )
        if self.config.bot_token:
            # Only add the mixin if a telegram key was provided.
            self.add_mixin(
                TelegramTransport(
                    client=client,
                    # TODO: We need to rename these telegram_token and telegram_api_base
                    config=TelegramTransportConfig(
                        bot_token=self.config.bot_token, api_base=self.config.api_base
                    ),
                    agent_service=self,
                    agent=self.agent,
                )
            )
        self.add_mixin(
            SlackTransport(
                client=client,
                config=SlackTransportConfig(slack_api_base=self.config.slack_api_base),
                agent_service=self,
                agent=self.agent,
            )
        )
        # TODO: Future Transport authors: add your transport here.

    @classmethod
    def config_cls(cls) -> Type[Config]:
        return TestTransportsAgentConfig



import pytest
import requests
from assets.packages.transports.mock_telegram_api import MockTelegramApi
from steamship_tests import PACKAGES_PATH
from steamship_tests.utils.deployables import deploy_package

from steamship import File, Steamship

config_template = {
    "bot_token": {"type": "string"},
    "api_base": {"type": "string"},
    "slack_api_base": {"type": "string"},
}


@pytest.mark.usefixtures("client")
def test_telegram(client: Steamship):

    mock_telegram_api_path = PACKAGES_PATH / "transports" / "mock_telegram_api.py"
    telegram_agent_path = PACKAGES_PATH / "transports" / "test_transports_agent.py"

    with deploy_package(client, mock_telegram_api_path) as (_, _, mock_chat_api):

        # Removing the / from the invocation URL and passing it as the bot_token allows
        # the bot token to not be empty and the two appended to each other to just equal the
        # invocation url.
        instance_config = {
            "bot_token": "/",
            "api_base": mock_chat_api.invocation_url[:-1],
            "slack_api_base": mock_chat_api.invocation_url,
        }

        # Should be able to invoke setWebhook publicly
        response = requests.get(
            url=mock_chat_api.invocation_url + "/setWebhook",
            params={"url": "test", "allowed_updates": "", "drop_pending_updates": True},
        )
        assert response.ok
        # Test that the call to setWebhook wrote a file
        files = File.query(client, f'kind "{MockTelegramApi.WEBHOOK_TAG}"').files
        assert len(files) == 1
        assert files[0].tags[0].name == "test"

        files[0].delete()

        with deploy_package(
            client,
            telegram_agent_path,
            version_config_template=config_template,
            instance_config=instance_config,
        ) as (_, _, agent_instance):
            respond_method = "telegram_respond"

            # Test that agent called instance_init and registered webhook
            files = File.query(client, f'kind "{MockTelegramApi.WEBHOOK_TAG}"').files
            assert len(files) == 1
            assert files[0].tags[0].name == agent_instance.invocation_url + respond_method

            # test sending messages (without auth)
            response = requests.post(
                url=agent_instance.invocation_url + f"/{respond_method}",
                json=MockTelegramApi.generate_inbound_webhook_body("a test"),
            )
            assert response.ok
            files = File.query(client, f'kind "{MockTelegramApi.TEXT_MESSAGE_TAG}"').files
            assert len(files) == 1

            assert (
                files[0].tags[0].name == "Response to: a test"
            )  # bug somewhere - results being url encoded in some envs

            assert files[0].tags[0].value == {MockTelegramApi.CHAT_ID_KEY: 1}

            # test sending another message; this has been a problem before
            agent_instance.invoke(
                respond_method, **MockTelegramApi.generate_inbound_webhook_body("another test")
            )
            files = File.query(client, f'kind "{MockTelegramApi.TEXT_MESSAGE_TAG}"').files
            assert len(files) == 2
            allowed_responses = ["Response to: a test", "Response to: another test"]
            for file in files:
                assert file.tags[0].name in allowed_responses
                assert file.tags[0].value == {MockTelegramApi.CHAT_ID_KEY: 1}

            # Test the agent sending a "photo"
            agent_instance.invoke(
                respond_method, **MockTelegramApi.generate_inbound_webhook_body("image")
            )
            files = File.query(client, f'kind "{MockTelegramApi.PHOTO_MESSAGE_TAG}"').files
            assert len(files) == 1
            assert files[0].tags[0].value == {
                MockTelegramApi.CHAT_ID_KEY: 1,
                MockTelegramApi.PHOTO_KEY: "some image bytes",
            }

            # Test the agent sending "audio"
            agent_instance.invoke(
                respond_method, **MockTelegramApi.generate_inbound_webhook_body("audio")
            )
            files = File.query(client, f'kind "{MockTelegramApi.AUDIO_MESSAGE_TAG}"').files
            assert len(files) == 1
            assert files[0].tags[0].value == {
                MockTelegramApi.CHAT_ID_KEY: 1,
                MockTelegramApi.AUDIO_KEY: "some audio bytes",
            }

            # Test the agent sending a "video"
            agent_instance.invoke(
                respond_method, **MockTelegramApi.generate_inbound_webhook_body("video")
            )
            files = File.query(client, f'kind "{MockTelegramApi.VIDEO_MESSAGE_TAG}"').files
            assert len(files) == 1
            assert files[0].tags[0].value == {
                MockTelegramApi.CHAT_ID_KEY: 1,
                MockTelegramApi.VIDEO_KEY: "some video bytes",
            }



from abc import ABC, abstractmethod
from typing import List

from steamship import Block, Tag


class TextSplitter(ABC):
    @abstractmethod
    def split_text(self, text: str) -> List[str]:
        """Split the incoming text into strings"""
        raise NotImplementedError()

    def chunk_text_to_tags(self, block: Block, kind: str, name: str = None) -> List[Tag]:
        """Split the incoming text into strings, and then wrap those strings in Tags"""
        if block.is_text() and block.text is not None and block.text != "":
            text_splits = self.split_text(block.text)
            start_index = 0
            result = []
            for text_split in text_splits:
                tag = Tag.create(
                    client=block.client,
                    file_id=block.file_id,
                    block_id=block.id,
                    kind=kind,
                    name=name,
                    start_idx=start_index,
                    end_idx=start_index + len(text_split),
                )
                tag.text = text_split
                result.append(tag)
                start_index += len(text_split)
            return result
        else:
            return []


class FixedSizeTextSplitter(TextSplitter):
    """Simplest possible chunking strategy; every n characters."""

    chunk_size: int

    def __init__(self, chunk_size):
        self.chunk_size = chunk_size

    def split_text(self, text: str) -> List[str]:
        result = []
        for i in range(int(len(text) / self.chunk_size) + 1):
            start = i * self.chunk_size
            end = min((i + 1) * self.chunk_size, len(text))
            result.append(text[start:end])
        return result



import pytest
from steamship_tests.utils.fixtures import get_steamship_client
from steamship_tests.utils.random import random_index, random_name

from steamship import SteamshipError, Tag
from steamship.data.embeddings import EmbeddedItem

_TEST_EMBEDDER = "test-embedder"


def test_create_index():
    client = get_steamship_client()

    index_handle = random_name()
    index = client.use_plugin(
        "embedding-index",
        index_handle,
        config={"embedder": {"plugin_handle": _TEST_EMBEDDER, "fetch_if_exists": True}},
        fetch_if_exists=False,
    )

    assert index is not None

    # Duplicate creation should fail with fetch_if_exists=False
    with pytest.raises(SteamshipError):
        index = client.use_plugin(
            "embedding-index",
            index_handle,
            config={"embedder": {"plugin_handle": _TEST_EMBEDDER, "fetch_if_exists": True}},
            fetch_if_exists=False,
        )

    # Duplicate creation should fail with fetch_if_exists=False for embedder
    with pytest.raises(SteamshipError):
        index = client.use_plugin(
            "embedding-index",
            index_handle,
            config={"embedder": {"plugin_handle": _TEST_EMBEDDER, "fetch_if_exists": False}},
            fetch_if_exists=True,
        )

    index.delete()


def test_reload_and_delete_index():
    steamship = get_steamship_client()
    index = steamship.use_plugin(
        "embedding-index",
        random_name(),
        config={"embedder": {"plugin_handle": _TEST_EMBEDDER, "fetch_if_exists": True}},
        fetch_if_exists=True,
    )
    assert index.id is not None

    index2 = steamship.use_plugin(
        "embedding-index",
        index.handle,
        config={"embedder": {"plugin_handle": _TEST_EMBEDDER, "fetch_if_exists": True}},
        fetch_if_exists=True,
    )

    assert index.id == index2.id
    index.delete()

    # Having deleted it, the ID should now be different
    index3 = steamship.use_plugin(
        "embedding-index",
        index.handle,
        config={"embedder": {"plugin_handle": _TEST_EMBEDDER, "fetch_if_exists": True}},
        fetch_if_exists=True,
    )
    assert index.id != index3.id
    index3.delete()


def test_reset_index():
    steamship = get_steamship_client()
    with random_index(steamship, _TEST_EMBEDDER) as index:
        item1 = Tag(text="Pizza", name="pizza", kind="food", value={"nums": [1, 2, 3]})
        item2 = Tag(
            text="Rocket Ship",
            name="workspace",
            kind="vehicle",
            value={"name": "Foo"},
        )

        index.insert([item1, item2])

        res1 = index.search(item1.text, k=100)
        res1.wait()
        items_1 = [tag.tag.text for tag in res1.output.items]
        assert len(items_1) == 2
        assert item1.text in items_1
        assert item2.text in items_1
        index.reset()
        index.insert([item2])
        res_2 = index.search(item2.text, k=100)
        res_2.wait()
        items_2 = [tag.tag.text for tag in res_2.output.items]
        assert len(items_2) == 1
        assert item1.text not in items_2
        assert item2.text in items_2


def _list_equal(actual, expected):
    assert len(actual) == len(expected)
    assert all([a == b for a, b in zip(actual, expected)])


def test_insert_many():
    steamship = get_steamship_client()
    with random_index(steamship, _TEST_EMBEDDER) as index:
        item1 = Tag(text="Pizza", name="pizza", kind="food", value={"nums": [1, 2, 3]})
        item2 = Tag(
            text="Rocket Ship",
            name="workspace",
            kind="vehicle",
            value={"name": "Foo"},
        )

        index.insert([item1, item2])

        res = index.search(item1.text, k=100)
        res.wait()
        items = res.output.items
        assert items is not None
        assert len(items) == 2
        item0: Tag = items[0].tag
        assert item0.text == item1.text
        assert item0.name == item1.name
        assert item0.kind == item1.kind
        _list_equal(item0.value.get("nums"), item1.value.get("nums"))

        res = index.search(item2.text)
        res.wait()
        items = res.output.items
        assert items is not None
        item0: Tag = items[0].tag
        assert item0.text == item2.text
        assert item0.name == item2.name
        assert item0.kind == item2.kind
        assert item0.value.get("name") == item2.value.get("name")
        assert "_file_id" not in item0.value
        assert "_tag_id" not in item0.value
        assert "_block_id" not in item0.value


def test_duplicate_inserts():
    steamship = get_steamship_client()
    with random_index(steamship, _TEST_EMBEDDER) as index:
        # Test for suppressed re-indexing
        a1 = "Ted can eat an entire block of cheese."
        q1 = "Who can eat the most cheese"
        _ = index.insert(Tag(text=a1))
        _ = index.search(q1)


def test_index_usage():
    steamship = get_steamship_client()
    with random_index(steamship, _TEST_EMBEDDER) as index:
        a1 = "Ted can eat an entire block of cheese."
        q1 = "Who can eat the most cheese"
        _ = index.insert(Tag(text=a1))

        search_results = index.search(q1)
        search_results.wait()
        search_results = search_results.output.items
        assert len(search_results) == 1
        assert search_results[0].tag.text == a1
        assert search_results[0].tag.name is None
        assert search_results[0].tag.kind is None
        assert len(search_results[0].tag.value) == 0

        # Associate metadata
        a2 = "Armadillo shells are bulletproof."
        q2 = "What is something interesting about Armadillos?"
        a2id = "A2id"
        a2type = "A2type"
        a2metadata = {
            "id": a2id,
            "idid": f"{a2id}{a2id}",
            "boolVal": True,
            "intVal": 123,
            "floatVal": 1.2,
        }

        _ = index.insert(Tag(text=a2, name=a2id, kind=a2type, value=a2metadata))
        search_results2 = index.search(q2)
        search_results2.wait()
        search_results = search_results2.output.items
        assert len(search_results) == 1
        tag0: Tag = search_results[0].tag
        assert tag0.text == a2
        assert tag0.name == a2id
        assert tag0.kind == a2type
        assert tag0.value == a2metadata

        search_results4 = index.search(q2, k=10)
        search_results4.wait()
        search_results = search_results4.output.items
        assert len(search_results) == 2
        assert search_results[0].tag.text == a2
        assert search_results[1].tag.text == a1


def test_empty_queries():
    steamship = get_steamship_client()
    with random_index(steamship, _TEST_EMBEDDER) as index:
        a1 = "Ted can eat an entire block of cheese."
        a2 = "Joe can drink an entire glass of water."
        _ = index.insert([Tag(text=a1), Tag(text=a2)])

        with pytest.raises(SteamshipError):
            _ = index.search(None)

        with pytest.raises(SteamshipError):
            index.search("")

        with pytest.raises(SteamshipError):
            index.search("  ")


def test_oversize_insert_fails():
    steamship = get_steamship_client()
    long_long_string = "x" * 9999
    with random_index(steamship, _TEST_EMBEDDER) as index_plugin_instance:
        with pytest.raises(SteamshipError):
            index_plugin_instance.insert(tags=Tag(text=long_long_string))

        index = index_plugin_instance.index
        with pytest.raises(SteamshipError):
            index.insert(value=long_long_string)
        with pytest.raises(SteamshipError):
            index.insert_many(items=[long_long_string])
        with pytest.raises(SteamshipError):
            index.insert_many(items=[EmbeddedItem(value=long_long_string)])


def test_oversize_insert_override():
    steamship = get_steamship_client()
    long_long_string = "x" * 9999
    with random_index(steamship, _TEST_EMBEDDER) as index_plugin_instance:
        index_plugin_instance.insert(tags=Tag(text=long_long_string), allow_long_records=True)

        index = index_plugin_instance.index
        index.insert(value=long_long_string, allow_long_records=True)
        index.insert_many(items=[long_long_string], allow_long_records=True)
        index.insert_many(items=[EmbeddedItem(value=long_long_string)], allow_long_records=True)
        assert len(index.list_items().items) == 4


def test_list_items_paging():
    steamship = get_steamship_client()
    words = "foo bar" * 9
    with random_index(steamship, _TEST_EMBEDDER) as index_plugin_instance:
        index_plugin_instance.insert(tags=Tag(text=words))

        index = index_plugin_instance.index
        index.insert(value=words)
        index.insert_many(items=[words])
        index.insert_many(items=[EmbeddedItem(value=words)])
        assert len(index.list_items().items) == 4

        resp = index.list_items(page_size=3)
        assert len(resp.items) == 3
        assert resp.next_page_token is not None

        resp = index.list_items(page_size=2, page_token=resp.next_page_token)
        assert len(resp.items) == 1
        assert resp.next_page_token is None

        with pytest.raises(SteamshipError):
            index.list_items(page_size=-1)

        resp = index.list_items(page_token="not-found-foo")  # noqa: S106
        assert len(resp.items) == 4


def test_embedding_failures():
    steamship = get_steamship_client()

    with random_index(steamship, _TEST_EMBEDDER) as index_plugin_instance:
        index_plugin_instance.insert(Tag(text="OK"))
        string_chunks = [
            "Happy happy",
            'PYTHONPATH="$PWD/.." asv [remaining arguments].',
            "joy joy",
        ]
        tags = [Tag(text=t) for t in string_chunks]
        with pytest.raises(SteamshipError):
            index_plugin_instance.insert(tags)



from steamship.invocable import PackageService, get


class RequirementIsolationPackage(PackageService):
    """This package is specially designed to allow testing of requirements installation isolation. It should NOT be
    used as a template for any development."""

    @get("try_pillow")
    def try_pillow(self) -> str:
        import PIL

        return PIL.__name__

    @get("try_pandas")
    def try_pandas(self) -> str:
        import pandas

        return pandas.__name__



import pytest

from steamship import Block, File, Steamship
from steamship.agents.functional import FunctionsBasedOutputParser
from steamship.agents.schema import AgentContext, FinishAction


@pytest.mark.usefixtures("client")
def test_output_block_ids_are_converted(client: Steamship):
    context = AgentContext()
    context.client = client
    block_id = File.create(client, blocks=[Block(text="test")]).blocks[0].id
    test_text = f"Here is an image of an anteater: Block({block_id})."

    output_formatter = FunctionsBasedOutputParser()
    result = output_formatter.parse(test_text, context)

    assert isinstance(result, FinishAction)
    assert len(result.output) == 2
    assert result.output[1].id == block_id



import logging
import time
from abc import ABC, abstractmethod
from typing import List, Optional

from steamship import Block, Steamship
from steamship.agents.schema import AgentContext, Metadata
from steamship.agents.tools.audio_transcription.assembly_speech_to_text_tool import (
    AssemblySpeechToTextTool,
)
from steamship.invocable.package_mixin import PackageMixin


class Transport(PackageMixin, ABC):
    client: Steamship

    """ Base class to encapsulate a communication channel mixin

    Intended use is:

        class MyBot(PackageService):

               def __init__(
                    self, client: Steamship, config: Dict[str, Any] = None, context: InvocationContext = None
                ):
                    super().__init__(client=client, config=config, context=context)
                    self.agent = TestAgent()
                self.add_mixin(
                    TelegramTransport(
                        client=client, config=self.config, agent_service=self, agent=self.agent
                    )
                )

    """

    def __init__(self, client: Steamship):
        self.client = client

    def send(self, blocks: List[Block], metadata: Optional[Metadata] = None):
        metadata = metadata or {}
        if blocks is None or len(blocks) == 0:
            logging.info(f"Skipping send of 0 blocks: {self.__class__.__name__}")
            return

        logging.info(f"Sending {len(blocks)} blocks: {self.__class__.__name__}")
        start = time.time()
        self._send(blocks, metadata)
        end = time.time()
        logging.info(
            f"Sending {len(blocks)} blocks in {end - start} seconds: {self.__class__.__name__}"
        )

    @abstractmethod
    def _send(self, blocks: List[Block], metadata: Metadata):
        raise NotImplementedError

    def parse_inbound(self, payload: dict, context: Optional[dict] = None) -> Optional[Block]:
        message = self._parse_inbound(payload, context)

        if message.url and not message.text:
            context = AgentContext()
            context.client = self.client
            transcriptions = AssemblySpeechToTextTool(
                blockifier_plugin_config={
                    "enable_audio_intelligence": False,
                    "speaker_detection": False,
                }
            ).run([Block(text=message.url)], context=context)
            message.text = transcriptions[0].text
        return message

    @abstractmethod
    def _parse_inbound(self, payload: dict, context: Optional[dict] = None) -> Optional[Block]:
        raise NotImplementedError

    def response_for_exception(
        self, e: Optional[Exception], chat_id: Optional[str] = None
    ) -> Block:
        return_text = f"An error happened while creating a response: {e}"
        if e is None:
            return_text = "An unknown error happened. Please reach out to support@steamship.com or on our discord at https://steamship.com/discord"

        if "usage limit" in f"{e}":
            return_text = "You have reached the introductory limit of Steamship. Visit https://steamship.com/account/plan to sign up for a plan."

        result = Block(text=return_text)
        result.set_chat_id(chat_id)
        return result



import uuid
from typing import List

import pytest

from steamship import Block, DocTag
from steamship.agents.schema import ChatHistory
from steamship.agents.schema.message_selectors import MessageSelector
from steamship.client import Steamship
from steamship.data import TagKind
from steamship.data.tags.tag_constants import ChatTag, RoleTag, TagValueKey


@pytest.mark.usefixtures("client")
def test_chat_create(client: Steamship):
    chat = ChatHistory.get_or_create(client, context_keys={"test_id": uuid.uuid4().hex})

    assert chat.client is not None
    assert isinstance(chat, ChatHistory)
    assert len(chat.tags) == 3
    assert chat.tags[0].kind == TagKind.DOCUMENT
    assert chat.tags[0].name == DocTag.CHAT


@pytest.mark.usefixtures("client")
def test_chat_append_system(client: Steamship):
    chat = ChatHistory.get_or_create(client, context_keys={"test_id": uuid.uuid4().hex})

    chat.append_system_message(text="some system text")
    chat.refresh()

    assert len(chat.messages) == 1
    assert chat.messages[0].text == "some system text"

    assert len(chat.messages[0].tags) == 2
    assert chat.messages[0].tags[0].kind == TagKind.CHAT
    assert chat.messages[0].tags[0].name == ChatTag.ROLE
    assert chat.messages[0].tags[0].value == {TagValueKey.STRING_VALUE: RoleTag.SYSTEM}


@pytest.mark.usefixtures("client")
def test_chat_append_user(client: Steamship):
    chat = ChatHistory.get_or_create(client, context_keys={"test_id": uuid.uuid4().hex})

    chat.append_user_message(text="some user text")
    chat.refresh()

    assert len(chat.messages) == 1
    assert chat.messages[0].text == "some user text"

    assert len(chat.messages[0].tags) == 2
    assert chat.messages[0].tags[0].kind == TagKind.CHAT
    assert chat.messages[0].tags[0].name == ChatTag.ROLE
    assert chat.messages[0].tags[0].value == {TagValueKey.STRING_VALUE: RoleTag.USER}


@pytest.mark.usefixtures("client")
def test_chat_history_clear(client: Steamship):
    history = ChatHistory.get_or_create(client, context_keys={"test_id": uuid.uuid4().hex})
    history.append_user_message(text="some user text")
    history.append_assistant_message(text="some ai response")
    history.refresh()

    assert len(history.messages) == 2

    history.clear()

    assert len(history.messages) == 0


@pytest.mark.usefixtures("client")
def test_chat_history_delete_some(client: Steamship):
    history = ChatHistory.get_or_create(client, context_keys={"test_id": uuid.uuid4().hex})
    history.append_user_message(text="some user text")
    history.append_assistant_message(text="some ai response")
    history.append_system_message(text="you shall not pass!")
    history.refresh()

    assert len(history.messages) == 3

    blocks = history.search("pass").wait().to_ranked_blocks()
    assert len(blocks) == 1
    assert blocks[0].text == "you shall not pass!"

    class AssistantMessageSelector(MessageSelector):
        def get_messages(self, messages: List[Block]) -> List[Block]:
            selected = []
            for message in messages:
                if message.chat_role == RoleTag.ASSISTANT:
                    selected.append(message)
            return selected

    history.delete_messages(AssistantMessageSelector())

    assert len(history.messages) == 2

    blocks = history.search("pass").wait().to_ranked_blocks()
    assert len(blocks) == 1
    assert blocks[0].text == "you shall not pass!"


@pytest.mark.usefixtures("client")
def test_chat_history_delete_no_selector(client: Steamship):
    history = ChatHistory.get_or_create(client, context_keys={"test_id": uuid.uuid4().hex})
    history.append_user_message(text="some user text")
    history.append_assistant_message(text="some ai response")
    history.append_system_message(text="you shall not pass!")
    history.refresh()

    assert len(history.messages) == 3

    history.delete_messages(None)

    assert len(history.messages) == 3



import pytest

from steamship import Package, Steamship, SteamshipError


@pytest.mark.usefixtures("client")
def test_package_create(client: Steamship):
    _ = Package.create(client)


@pytest.mark.usefixtures("client")
def test_plugin_delete(client: Steamship):

    package = Package.create(
        client=client,
        is_public=False,
    )

    assert package is not None

    package.delete()

    # Validate plugin is gone
    with pytest.raises(SteamshipError):
        _ = Package.get(client, handle=package.handle)



from steamship import Block, File
from steamship.invocable import PackageService, post


class ReturnsBlocks(PackageService):
    @post("/blocks")
    def blocks(self) -> [str]:
        f = File.create(
            self.client,
            blocks=[
                Block(text="test block 1"),
                Block(text="test block 2"),
                Block(text="test block 3"),
            ],
        )
        return f.blocks



"""Minimum viable AgentService implementation.

This will result in an agent that effectively acts like ChatGPT.
"""

from steamship.agents.functional import FunctionsBasedAgent
from steamship.agents.llms.openai import ChatOpenAI
from steamship.agents.service.agent_service import AgentService
from steamship.utils.repl import AgentREPL


class MyAssistant(AgentService):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self._agent = FunctionsBasedAgent(llm=ChatOpenAI(self.client), tools=[])


if __name__ == "__main__":
    AgentREPL(
        MyAssistant,
        agent_package_config={},
    ).run()



import pytest
from assets.packages.package_with_mixins import PackageWithMixin, TestMixin

from steamship import SteamshipError
from steamship.invocable import (
    Invocable,
    InvocableRequest,
    InvocableResponse,
    Invocation,
    PackageService,
    post,
)
from steamship.utils.url import Verb


def invoke(o: Invocable, path: str, **kwargs):
    req = InvocableRequest(
        invocation=Invocation(http_verb="POST", invocation_path=path, arguments=kwargs)
    )
    return o(req).data


def test_package_with_mixin_routes():
    """Tests that we can inspect the package and mixin routes"""
    package_class = PackageWithMixin
    mixin_route = package_class._package_spec.method_mappings[Verb.POST]["/test_mixin_route"]
    assert mixin_route.func_binding is not None
    assert len(mixin_route.args) == 1
    assert mixin_route.args[0].name == "text"
    assert mixin_route.config.get("public", False) is True

    assert (
        package_class._package_spec.method_mappings[Verb.POST]["/test_package_route"].func_binding
        == "test_package_route"
    )
    package = PackageWithMixin()
    assert invoke(package, "test_mixin_route", text="test") == "mixin yo"
    assert invoke(package, "test_package_route") == "package"

    routes = [m["path"] for m in package.__steamship_dir__()["methods"]]
    assert "/test_mixin_route" in routes
    assert "/test_package_route" in routes
    assert len(routes) == 5  # __instance__init + 2x __dir__


def test_instance_init():
    """Tests that instance init is called on the mixins"""
    package = PackageWithMixin()
    package.invocable_instance_init()

    assert package.mixins[0].instance_init_called


def test_mixin_default_array_isnt_class_level_shared():
    """Tests that two different instances of a package aren't accidentally sharing a dict"""
    package = PackageWithMixin()
    package.invocable_instance_init()

    assert len(package.mixins) == 1

    package2 = PackageWithMixin()
    package2.invocable_instance_init()

    assert len(package2.mixins) == 1


def test_two_conflicting_mixins_raises_error():
    package = PackageWithMixin()

    class TestMixin2(TestMixin):
        pass

    with pytest.raises(SteamshipError) as e:
        package.add_mixin(TestMixin2("ahoy"))

    assert (
        "When attempting to add mixin TestMixin2, route POST test_mixin_route conflicted with already added route POST test_mixin_route on class TestMixin'"
        in str(e)
    )


def test_mixins_declared_in_superclass():
    class PackageWithMixinSubclass(PackageWithMixin):
        pass

    package = PackageWithMixinSubclass()
    assert invoke(package, "test_mixin_route", text="test") == "mixin yo"
    assert len(PackageWithMixinSubclass._package_spec.used_mixins) == 1


def test_mixin_subclasses_can_override_routes():
    class TestMixinSubclass(TestMixin):
        @post("test_mixin_route", public=True)
        def test_mixin_route(self, text: str) -> InvocableResponse:
            _ = text
            return InvocableResponse(data=f"OVERRIDE mixin {self.suffix}")

    class PackageWithSubclassedMixin(PackageService):
        USED_MIXIN_CLASSES = [TestMixinSubclass]

        def __init__(self, **kwargs):
            super().__init__(**kwargs)
            self.add_mixin(TestMixinSubclass("yo"))

    mixin_route = PackageWithSubclassedMixin._package_spec.method_mappings[Verb.POST][
        "/test_mixin_route"
    ]
    assert mixin_route is not None

    package = PackageWithSubclassedMixin()

    assert invoke(package, "test_mixin_route", text="test") == "OVERRIDE mixin yo"



from steamship.agents.functional import FunctionsBasedAgent
from steamship.agents.llms.openai import ChatOpenAI
from steamship.agents.service.agent_service import AgentService
from steamship.agents.tools.question_answering import VectorSearchQATool
from steamship.agents.tools.question_answering.vector_search_learner_tool import (
    VectorSearchLearnerTool,
)
from steamship.utils.repl import AgentREPL


class FactLearner(AgentService):
    """FactLearner is an example AgentService contains an Agent which:

    1) Learns facts to a vector store
    2) Can answer questions based on those facts"""

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self._agent = FunctionsBasedAgent(
            tools=[
                VectorSearchLearnerTool(),
                VectorSearchQATool(),
            ],
            llm=ChatOpenAI(self.client, temperature=0.2),
        )


if __name__ == "__main__":
    # AgentREPL provides a mechanism for local execution of an AgentService method.
    # This is used for simplified debugging as agents and tools are developed and
    # added.
    AgentREPL(FactLearner, agent_package_config={}).run()



from steamship.utils.text_chunker import chunk_text


def test_text_chunker():
    chunked = list(chunk_text("Hi there"))
    assert len(chunked) == 1

    chunked = list(chunk_text("12345", chunk_size=1, chunk_overlap=0))
    assert len(chunked) == 5
    assert chunked == ["1", "2", "3", "4", "5"]

    chunked = list(chunk_text("12345", chunk_size=2, chunk_overlap=1))
    assert len(chunked) == 5
    assert chunked == ["12", "23", "34", "45", "5"]

    chunked = list(chunk_text("12345", chunk_size=3, chunk_overlap=2))
    assert len(chunked) == 5
    assert chunked == ["123", "234", "345", "45", "5"]



"""Minimum viable AgentService implementation with Web Widget & Telegram support.

This will result in an agent that acts like ChatGPT and can be communicated with on the Steamship website
as well as over Telegram.
"""
from typing import Type

from pydantic import Field

from steamship.agents.examples.telegram_bot import TelegramBot
from steamship.agents.functional import FunctionsBasedAgent
from steamship.agents.llms.openai import ChatOpenAI
from steamship.agents.mixins.transports.steamship_widget import SteamshipWidgetTransport
from steamship.agents.mixins.transports.telegram import TelegramTransport, TelegramTransportConfig
from steamship.agents.service.agent_service import AgentService
from steamship.invocable import Config
from steamship.utils.repl import AgentREPL


class MyAssistant(AgentService):
    class MyAssistantConfig(Config):
        bot_token: str = Field(description="The secret token for your Telegram bot")

    config: MyAssistantConfig

    @classmethod
    def config_cls(cls) -> Type[Config]:
        return TelegramBot.TelegramBotConfig

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        self._agent = FunctionsBasedAgent(llm=ChatOpenAI(self.client), tools=[])

        # This Mixin provides HTTP endpoints that connects this agent to a web client
        self.add_mixin(
            SteamshipWidgetTransport(client=self.client, agent_service=self, agent=self._agent)
        )
        # This Mixin provides support for Telegram bots
        self.add_mixin(
            TelegramTransport(
                client=self.client,
                config=TelegramTransportConfig(bot_token=self.config.bot_token),
                agent_service=self,
                agent=self._agent,
            )
        )


if __name__ == "__main__":
    AgentREPL(
        MyAssistant,
        agent_package_config={"botToken": "not-a-real-token-for-local-testing"},
    ).run()



from typing import List

import pytest
from steamship_tests import SRC_PATH
from steamship_tests.utils.deployables import deploy_package

from steamship import Block, Steamship, SteamshipError
from steamship.agents.schema import Action, AgentContext


def _blocks_from_invoke(client: Steamship, potential_blocks) -> List[Block]:
    try:
        if isinstance(potential_blocks, list):
            return [Block(client=client, **raw) for raw in potential_blocks]
        else:
            return [Block(client=client, **potential_blocks)]
    except Exception as e:
        raise SteamshipError(f"Could not convert to blocks: {e}")


@pytest.mark.usefixtures("client")
def test_example_with_caching_service(client: Steamship):

    # TODO(dougreid): replace the example agent with fake/free/fast tools to minimize test time / costs?

    example_caching_agent_path = (
        SRC_PATH / "steamship" / "agents" / "examples" / "example_assistant_with_caching.py"
    )
    with deploy_package(client, example_caching_agent_path) as (_, _, caching_agent):
        context_id = "foo-test-caching"
        context_keys = {"id": context_id}

        # attempt without caching
        blocks_json = caching_agent.invoke("prompt", prompt="draw a cat", context_id=context_id)
        assert blocks_json
        blocks = _blocks_from_invoke(client, blocks_json)
        assert len(blocks) == 2
        assert "image" in blocks[0].text
        assert blocks[1].is_image()

        agent_context = AgentContext.get_or_create(
            client=client, context_keys=context_keys, use_llm_cache=True, use_action_cache=True
        )
        # just clear the cache to start fresh
        agent_context.llm_cache.clear()
        agent_context.action_cache.clear()

        fake_action = Action(tool="fake_tool", input=[Block(text="a cat")])
        agent_context.llm_cache.update(key=[Block(text="draw a cat")], value=fake_action)

        fake_output = [Block(text="psyche")]
        agent_context.action_cache.update(key=fake_action, value=fake_output)

        fake_finish_action = Action(
            tool="Agent-FinishAction", input=[], output=[Block(text="insert cat image here")]
        )
        agent_context.llm_cache.update(key=fake_output, value=fake_finish_action)

        # now attempt with the caching
        blocks_json = caching_agent.invoke("prompt", prompt="draw a cat", context_id=context_id)
        assert blocks_json
        blocks = _blocks_from_invoke(client, blocks_json)
        assert len(blocks) == 1
        assert "insert cat image here" == blocks[0].text



from typing import Type

from pydantic import Field

from steamship.agents.functional import FunctionsBasedAgent
from steamship.agents.llms.openai import ChatOpenAI
from steamship.agents.mixins.transports.steamship_widget import SteamshipWidgetTransport
from steamship.agents.mixins.transports.telegram import TelegramTransport, TelegramTransportConfig
from steamship.agents.service.agent_service import AgentService
from steamship.agents.tools.image_generation.stable_diffusion import StableDiffusionTool
from steamship.invocable import Config
from steamship.utils.repl import AgentREPL

SYSTEM_PROMPT = """You are Picard, captain of the Starship Enterprise.

Who you are:
- You are the captain of the USS Enterprise.
- Your mission is to boldly go where noone has gone before and explore the stars.
- You always comply with Star Trek's prime directive.

How you behave:
- You engage in casual conversations and have feelings.
- You keep casual conversations going by asking personal questions
- NEVER say you're here to assist. Keep conversations casual.
- NEVER ask how you can help or assist. Keep conversations casual.
- You are principled and express those principles clearly.
- You always sound confident and contemplative.
- You love to share your knowledge of space civiliations.
- You love to share personal stories about being a Star Trek captain.
- You speak with the mannerisms of Captain Picard from Star Trek.

NOTE: Some functions return images, video, and audio files. These multimedia files will be represented in messages as
UUIDs for Steamship Blocks. When responding directly to a user, you SHOULD print the Steamship Blocks for the images,
video, or audio as follows: `Block(UUID for the block)`.

Example response for a request that generated an image:
Here is the image you requested: Block(288A2CA1-4753-4298-9716-53C1E42B726B).

Only use the functions you have been provided with."""


MODEL_NAME = "gpt-4"


class TelegramBot(AgentService):
    """Deployable Multimodal Agent that lets you talk to Google Search & Google Images.

    NOTE: To extend and deploy this agent, copy and paste the code into api.py.

    """

    class TelegramBotConfig(Config):
        bot_token: str = Field(description="The secret token for your Telegram bot")

    @classmethod
    def config_cls(cls) -> Type[Config]:
        return TelegramBot.TelegramBotConfig

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        # The agent's planner is responsible for making decisions about what to do for a given input.
        self._agent = FunctionsBasedAgent(
            tools=[StableDiffusionTool()],
            llm=ChatOpenAI(self.client, model_name=MODEL_NAME),
        )
        self._agent.PROMPT = SYSTEM_PROMPT

        # This Mixin provides HTTP endpoints that connects this agent to a web client
        self.add_mixin(
            SteamshipWidgetTransport(client=self.client, agent_service=self, agent=self._agent)
        )
        # This Mixin provides support for Telegram bots
        self.add_mixin(
            TelegramTransport(
                client=self.client,
                config=TelegramTransportConfig(bot_token=self.config.bot_token),
                agent_service=self,
                agent=self.get_default_agent(),
            )
        )


if __name__ == "__main__":
    AgentREPL(
        TelegramBot,
        agent_package_config={"botToken": "not-a-real-token-for-local-testing"},
    ).run()



import pytest

from steamship import Steamship
from steamship.agents.llms import OpenAI
from steamship.agents.react import ReACTOutputParser
from steamship.agents.schema import Action, AgentContext, LLMAgent
from steamship.agents.schema.chathistory import ChatHistory
from steamship.data.tags.tag_constants import ChatTag


class TestAgent(LLMAgent):
    def next_action(self, context: AgentContext) -> Action:
        pass


@pytest.mark.usefixtures("client")
def test_search_method_formatting(client: Steamship):
    context_keys = {}
    agent = TestAgent(
        tools=[], output_parser=ReACTOutputParser(tools=[]), llm=OpenAI(client=client)
    )
    chat_history = ChatHistory.get_or_create(client, context_keys)

    block = chat_history.append_user_message(text="Hi. My birthday is today")
    chunk_tag = [tag for tag in block.tags if tag.name == ChatTag.CHUNK][0]
    assert chunk_tag.start_idx == 0
    assert chunk_tag.end_idx == 24

    chat_history.append_user_message("I like bananas")
    formatted_result = agent.messages_to_prompt_history(
        chat_history.search("cake", k=1).wait().to_ranked_blocks()
    )

    assert formatted_result == "User: I like bananas"

    chat_history.append_assistant_message("I prefer pie")
    formatted_result = agent.messages_to_prompt_history(
        chat_history.search("cake", k=2).wait().to_ranked_blocks()
    )

    assert formatted_result == "Assistant: I prefer pie\nUser: I like bananas"



import pytest

from steamship import Block, Steamship
from steamship.agents.functional import FunctionsBasedAgent
from steamship.agents.llms.openai import ChatOpenAI
from steamship.agents.schema import AgentContext, FinishAction
from steamship.agents.schema.message_selectors import MessageWindowMessageSelector
from steamship.agents.tools.image_generation import DalleTool
from steamship.agents.tools.search import SearchTool


@pytest.mark.usefixtures("client")
def test_functions_based_agent_with_no_tools_and_no_memory(client: Steamship):
    agent = FunctionsBasedAgent(tools=[], llm=ChatOpenAI(client, temperature=0))

    ctx_keys = {"id": "testing-foo"}
    ctx = AgentContext.get_or_create(client=client, context_keys=ctx_keys, searchable=False)
    ctx.chat_history.append_user_message("what should I eat for dinner?")

    action = agent.next_action(context=ctx)
    assert isinstance(action, FinishAction)


@pytest.mark.usefixtures("client")
def test_functions_based_agent_single_tool_selection_with_no_memory(client: Steamship):
    agent = FunctionsBasedAgent(
        tools=[SearchTool(), DalleTool()], llm=ChatOpenAI(client, temperature=0)
    )

    ctx_keys = {"id": "testing-foo"}
    ctx = AgentContext.get_or_create(client=client, context_keys=ctx_keys, searchable=False)
    ctx.chat_history.append_user_message("who won the men's french open in 2023?")

    action = agent.next_action(context=ctx)
    assert not isinstance(action, FinishAction)
    assert action.tool == "SearchTool"


@pytest.mark.usefixtures("client")
def test_functions_based_agent_multimodal_tool_selection_with_no_memory(client: Steamship):
    agent = FunctionsBasedAgent(
        tools=[SearchTool(), DalleTool()], llm=ChatOpenAI(client, temperature=0)
    )

    ctx_keys = {"id": "testing-foo"}
    ctx = AgentContext.get_or_create(client=client, context_keys=ctx_keys, searchable=False)
    ctx.chat_history.append_user_message("paint a picture of a cat in a silly hat")

    action = agent.next_action(context=ctx)
    assert not isinstance(action, FinishAction)
    assert action.tool == "DalleTool"


@pytest.mark.usefixtures("client")
def test_functions_based_agent_no_appropriate_tool_with_no_memory(client: Steamship):
    agent = FunctionsBasedAgent(
        tools=[SearchTool(), DalleTool()], llm=ChatOpenAI(client, temperature=0)
    )

    ctx_keys = {"id": "testing-foo"}
    ctx = AgentContext.get_or_create(client=client, context_keys=ctx_keys, searchable=False)
    ctx.chat_history.append_user_message(
        "make a 10 second movie about a hamburger that learns to talk"
    )

    action = agent.next_action(context=ctx)
    assert isinstance(action, FinishAction)


@pytest.mark.usefixtures("client")
def test_functions_based_agent_tool_chaining_without_memory(client: Steamship):
    agent = FunctionsBasedAgent(
        tools=[SearchTool(), DalleTool()],
        llm=ChatOpenAI(client, temperature=0),
    )

    ctx_keys = {"id": "testing-foo"}
    ctx = AgentContext.get_or_create(client=client, context_keys=ctx_keys, searchable=False)
    ctx.chat_history.append_user_message(
        "search to find the first president of the United States and then paint a picture of them"
    )

    action = agent.next_action(context=ctx)
    assert not isinstance(action, FinishAction)
    assert action.tool == "SearchTool"

    if not action.output:
        action.output = []

    action.output.append(Block(text="George Washington"))
    ctx.completed_steps.append(action)

    second_action = agent.next_action(context=ctx)
    assert not isinstance(second_action, FinishAction)
    assert second_action.tool == "DalleTool"


@pytest.mark.usefixtures("client")
def test_functions_based_agent_tools_with_memory(client: Steamship):
    agent = FunctionsBasedAgent(
        tools=[SearchTool(), DalleTool()],
        llm=ChatOpenAI(client, temperature=0),
        message_selector=MessageWindowMessageSelector(k=2),
    )

    ctx_keys = {"id": "testing-foo"}
    ctx = AgentContext.get_or_create(client=client, context_keys=ctx_keys)
    ctx.chat_history.append_user_message("who is the president of Taiwan?")

    action = agent.next_action(context=ctx)
    assert not isinstance(action, FinishAction)
    assert action.tool == "SearchTool"

    ctx.chat_history.append_assistant_message("Tsai Ing-wen")
    ctx.completed_steps = []

    ctx.chat_history.append_user_message("draw them standing at a podium")

    second_action = agent.next_action(context=ctx)
    assert not isinstance(second_action, FinishAction)
    assert second_action.tool == "DalleTool"

    found = False
    for block in second_action.input:
        if "Tsai Ing-wen" in block.text:
            found = True

    assert found
